#  ooo  oooo  oooo o   o     o    ooo   ooo  o   o  oooo ooooo o  ooo   oooo
# 0   0 0   0 0    00  0    0 0  0   0 0   0 0   0 0       0   0 0   0 0
# 0   0 oooo  oooo 0 0 0   0   0 0     0   0 0   0  oooo   0   0 0      oooo
# 0   0 o     0    0 o 0   ooooo 0   o 0   0 0   0      0  0   0 0   o      0
#  ooo  o     oooo o  0o  0     0 ooo   ooo   ooo  0oooo   0   0  ooo  0oooo
#
#  (  (  (  (  (  (  (  (  (  (  (  (  O  )  )  )  )  )  )  )  )  )  )  )  )
#
#    OpenAcoustics: The Open Source Collection for Computational Acoustics
#
#               Copyright (C) 2008-2012 The OpenAcoustics Team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


# Import required libraries
from numpy import array, dot
from scipy import pi, arccos, tan, sqrt, zeros
from scipy.linalg import norm
from scipy.sparse import lil_matrix
from scipy.sparse.linalg.dsolve import spsolve
from openacoustics.gmsh import *
from matplotlib import cm

# Define a function that returns cotangent of an angle t
def cot(t):
    return 1. / tan(t)

# Takes three node coordinates of a triangular element node0, node1,
# node2 and calculate elementary mass and stiffness matrices [M] and
# [K]
def triangularElement(node0, node1, node2):
    a = node1 - node0
    b = node2 - node1
    c = node0 - node2

    t0 = arccos(dot(-c, a) / (norm(c) * norm(a)))
    t1 = arccos(dot(-a, b) / (norm(a) * norm(b)))
    t2 = arccos(dot(-b, c) / (norm(b) * norm(c)))
    
    s = (norm(a) + norm(b) + norm(c)) / 2.
    S = sqrt(s * (s - norm(a)) * (s - norm(b)) * (s - norm(c)))

    K = array([[cot(t1) + cot(t2), -cot(t2), -cot(t1)],
               [-cot(t2), cot(t0) + cot(t2), -cot(t0)],
               [-cot(t1), -cot(t0), cot(t0) + cot(t1)]]) / 2.

    M = array([[2., 1., 1.],
               [1., 2., 1.],
               [1., 1., 2.]]) * S / 12.
    
    return K, M

geo = 'square.geo'

# Create an instance of class gmsh
gmsh = GMsh2D()

# Load the .geo file
gmsh.loadGeo(geo, 0.5, order=1)


# Load node indices and coordinates
nodes = gmsh.getNodes()
nodenum = nodes.shape[0] # Number of nodes

# Get the source node number
source = gmsh.getPoints("S")

# Load node numbers of triangular elements
elements = gmsh.getTriangles("", order=1)


# (x, y) coordiates of all nodes: (0, 0), (1, 0), (1, 1), (0, 1)
#nodes = array([[0.,0.], [1.,0.], [1.,1.], [0.,1.]])
# Node numbers that constitutes each of two triangular elements
# (connectivity)
#elements = array([[0,1,3], [1,2,3]])

# Get the number of nodes
numberOfNodes = nodes.shape[0]

# Create elementary matrices
K = lil_matrix((numberOfNodes, numberOfNodes))
M = lil_matrix((numberOfNodes, numberOfNodes))

# Calculate elementary matrices and assemble global system matrices
for element in elements:
    # Extract three node coordinates that constitute an element
    node0 = nodes[element[0],:]
    node1 = nodes[element[1],:]
    node2 = nodes[element[2],:]

    # Calculate elementary matrices
    Kn, Mn = triangularElement(node0, node1, node2)
    Mn /= (343. * 343.)

    # Assemble global system matrices
    for rowIndex in range(Kn.shape[0]):
        for columnIndex in range(Kn.shape[1]):
            K[element[rowIndex],element[columnIndex]] \
                += Kn[rowIndex,columnIndex]
            M[element[rowIndex],element[columnIndex]] \
                += Mn[rowIndex,columnIndex]

# Set a point source of unit intensity at node 0
q = zeros(numberOfNodes)
q[0] = 1.

# Calculate angular frequency from assumed stationary state at 170 Hz
omega = 2. * pi * 170.

# Solve the system equation
phi = spsolve(K - (omega ** 2) * M, q.T)
phi = abs(phi)

# Print the final solution
import matplotlib.collections as collections
verts=[]
val=[]
for element in elements:
    node0 = nodes[element[0],:]
    node1 = nodes[element[1],:]
    node2 = nodes[element[2],:]
    verts.append((node0,node1,node2))
    
    p0 = phi[element[0]]
    p1 = phi[element[1]]
    p2 = phi[element[2]]
    val.append((p0+p1+p2)/3.)

col = collections.PolyCollection(verts,cmap=cm.gray)
col.set_array(array(val))
fig=figure()
ax=fig.add_subplot(1,1,1,aspect='equal')
ax.add_collection(col)
fig.show()

