#  ooo  oooo  oooo o   o     o    ooo   ooo  o   o  oooo ooooo o  ooo   oooo
# 0   0 0   0 0    00  0    0 0  0   0 0   0 0   0 0       0   0 0   0 0
# 0   0 oooo  oooo 0 0 0   0   0 0     0   0 0   0  oooo   0   0 0      oooo
# 0   0 o     0    0 o 0   ooooo 0   o 0   0 0   0      0  0   0 0   o      0
#  ooo  o     oooo o  0o  0     0 ooo   ooo   ooo  0oooo   0   0  ooo  0oooo
#
#  (  (  (  (  (  (  (  (  (  (  (  (  O  )  )  )  )  )  )  )  )  )  )  )  )
#
#    OpenAcoustics: The Open Source Collection for Computational Acoustics
#
#               Copyright (C) 2008-2009 The OpenAcoustics Team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


__author__ = "t_ishiduka <http://sourceforge.net.users.t_ishiduka>" 
__date__ = "2010/08/19"
__credits__ = "ISHIZUKA Takashi"
__version__ = "$Revision: 1 $"

# import modules 
print 'Loading modules ...'
from scipy import array, linalg, zeros, dot, mgrid, pi, sqrt, special, real
from scipy.integrate import quadrature
from openacoustics import gmsh
import pylab


#=========================================================================
#                        Definitions of functions
#=========================================================================
# Hankel function of the first kind of order zero
def h01(x):
    return special.j0(x) + 1j * special.y0(x)


# Hankel function of the first kind of order one 
def h11(x):
    return special.j1(x) + 1j * special.y1(x)


# Green function of the 2-dimentional free field
# See eqn. (2.33)
def GreenFunction(receiver, source, waveNumber):
    r = receiver - source
    abs_r = sqrt(r[0] ** 2. + r[1] ** 2.)

    return 1j * h01(abs_r * waveNumber) / 4.


# Integrand of element integral Ia,
# on local axis xi.
# See eqn. (2.54)
def integrandIa(xi, element, receiver, waveNumber):
    r_x = receiver[0] - element['middlePoint'][0] + element['localAxis'][0] * xi
    r_y = receiver[1] - element['middlePoint'][1] + element['localAxis'][1] * xi
    abs_r = sqrt(r_x ** 2. + r_y ** 2.)
    dr_dn = - (r_x * element['normalVector'][0] + r_y * element['normalVector'][1]) \
            / abs_r

    return h11(abs_r * waveNumber) * dr_dn


# Integrand of element integral Ib,
# on local axis xi.
# See eqn. (2.55)
def integrandIb( xi, element, receiver, waveNumber ):
    r_x = receiver[0] - element['middlePoint'][0] + element['localAxis'][0] * xi
    r_y = receiver[1] - element['middlePoint'][1] + element['localAxis'][1] * xi
    abs_r = sqrt(r_x ** 2. + r_y ** 2.)

    return h01(abs_r * waveNumber)


# Singular integral of the Hankel function of the first kind of order zero
# Int_{0}^{a} h01(t) dt
# See eqn (2.39)
def singularIntegralH01(a):
    h01a = h01(a);
    h11a = h11(a);
    H0a = special.struve(0,a);
    H1a = special.struve(1,a);

    value = a * h01a + pi * a / 2. * (H0a * h11a - H1a * h01a);
    
    return value


#=========================================================================
#                              Main routine
#=========================================================================
#========================================================== Initialization 
print 'Initializing ...'

#-- sound speed -- assuming temperature = 15 in Celsius degree.
soundSpeed = 340.

#-- frequency --
frequency = 100.
angularFrequency = 2. * pi * frequency
waveNumber = angularFrequency / soundSpeed

#-- boundary elements  --
print '  Creating boundary elements...'

# set boundary conditions of lines
boundaryCondition = {'rigid':0., 'absorbing':1.}

# create mesher as instance of GMsh2D class
mesher = gmsh.GMsh2D()

# create elements using GMsh and get nodes of elements
mesher.loadGeo('sampleBemSquare.geo', 1)
node = mesher.getNodes()

# build element
element = []

# loop for boundary contidions, getting condition names.
for conditionName in boundaryCondition.keys():
    # get node indexes defining elements
    nodeIndex = mesher.getLines(conditionName)

    # set p1, p2 and admittance of each element and append to element
    for idx in nodeIndex:
        element.append({'p1':node[idx[0]], 'p2':node[idx[1]],
            'admittance':boundaryCondition[conditionName]})

nElement = len(element)

print '    The number of elements : ', nElement

#-- source & receivers --
source = array( [0.4, 0.4] )

print '  Source position : ', source

# mesh grid of recievers
print '  Creating mesh grid of recievers ...'

[rY, rX] = mgrid[0.45:-0.5:-0.1, -0.45:0.5:0.1]
(rk,rl) = rX.shape

print '    The number of recievers : ', rk*rl

#======================================= Preparation of element parameters 
for n in range(nElement):
    element[n]['middlePoint'] = (element[n]['p1'] + element[n]['p2']) / 2.
    element[n]['localAxis'] = (element[n]['p2'] - element[n]['p1']) / 2.
    element[n]['length'] = sqrt(element[n]['localAxis'][0] ** 2. \
            + element[n]['localAxis'][1] ** 2.) * 2.
    element[n]['normalVector'] = \
            array([ element[n]['localAxis'][1], -element[n]['localAxis'][0] ]) \
            * 2. / element[n]['length']

#============================================ Build coefficient matrix 'A'
print 'Carring out BEM ...'
print '  Building coefficient matrix \'A\' ...'

#See eqn. (2.24) and (2.22)
A = zeros((nElement,nElement), 'complex64')

# loop for boundary elements
j_4 = 1j / 4.
for n in range(nElement): 
    a = -1j * waveNumber * element[n]['admittance']
    halfLength = element[n]['length'] / 2.

    # loop for receivers at middle points of each element
    for m in range(nElement): 
        rm = element[m]['middlePoint']

        if m != n: # general components
            #See eqn. (2.54)
            Ia = - j_4 * waveNumber * halfLength \
                    * quadrature(integrandIa, -1, 1, args=(element[n],rm,waveNumber))[0]
            #See eqn. (2.55)
            Ib = j_4 * halfLength  \
                    * quadrature(integrandIb, -1, 1, args=(element[n],rm,waveNumber))[0]
            A[m,n] = Ia + a * Ib

        else: # diagonal components : take singular integral
            #See eqn. (2.46)
            Ib = (1j / (2. * waveNumber)) * \
                    singularIntegralH01(waveNumber * halfLength)
            A[m,n] = 1./2. + a * Ib

#=========================================== Build right hand vector 'psy'
print '  Building right hand vector \'psy\' ...'

#See eqn. (2.23) and (2.33)
psi = zeros((nElement,1), 'complex64')

for n in range(nElement):
    psi[n] = GreenFunction( element[n]['middlePoint'], source, waveNumber )

#===================================== Solve linear equation A * phi = psy
print '  Solving linear equation \'A * phi = psy\' ...'

#See eqn. (2.24)
phi = linalg.solve( A, psi )

#============================== Calculate velocity potentials at receivers
print '  Calculate velocity potentials at receivers ...'

#See eqn. (2.15) and (2.50)
phiR = zeros((rk, rl), 'complex64' )
An = zeros(nElement, 'complex64')

for k in range(rk):
    for l in range(rl):
        Rkl = array([rX[k,l], rY[k,l]])
    
        for n in range(nElement):
            a = -1j * waveNumber * element[n]['admittance']
            halfLength = element[n]['length'] / 2.
    
            #See eqn. (2.54)
            Ia = -j_4 * waveNumber * halfLength \
                    * quadrature(integrandIa, -1, 1, args=(element[n],Rkl,waveNumber))[0]
            #See eqn. (2.55)
            Ib = j_4 * halfLength \
                    * quadrature(integrandIb, -1, 1, args=(element[n],Rkl,waveNumber))[0]
            An[n] = Ia + a * Ib
    
        #See eqn. (2.50) 
        phiR[k,l] = GreenFunction(Rkl, source, waveNumber) - dot(An, phi)[0]

#============================================================ Show results
print 'Creating contour plot of Re(phiR) ...'

# plot contour
pylab.contourf(rX, rY, real(phiR), 64)
pylab.hold(True)

# plot boundary elements
# If the number of elements is too large, you should comment out these lines
nodeIndex = mesher.getLines('rigid')
for idx in nodeIndex:
    pylab.plot(node[idx][:,0], node[idx][:,1], '-bx')

nodeIndex = mesher.getLines('absorbing')
for idx in nodeIndex:
    pylab.plot(node[idx][:,0], node[idx][:,1], '-rx')

# plot source position
pylab.plot(array([source[0]]), array([source[1]]), 'bo')

# arrange graph view
pylab.axis('equal')
pylab.xlabel('x [m]')
pylab.ylabel('y [m]')
pylab.title('Real part of velocity potential')

# show graph
pylab.show()

print 'Done.'

# End






