#  ooo  oooo  oooo o   o     o    ooo   ooo  o   o  oooo ooooo o  ooo   oooo
# 0   0 0   0 0    00  0    0 0  0   0 0   0 0   0 0       0   0 0   0 0
# 0   0 oooo  oooo 0 0 0   0   0 0     0   0 0   0  oooo   0   0 0      oooo
# 0   0 o     0    0 o 0   ooooo 0   o 0   0 0   0      0  0   0 0   o      0
#  ooo  o     oooo o  0o  0     0 ooo   ooo   ooo  0oooo   0   0  ooo  0oooo
#
#  (  (  (  (  (  (  (  (  (  (  (  (  O  )  )  )  )  )  )  )  )  )  )  )  )
#
#    OpenAcoustics: The Open Source Collection for Computational Acoustics
#
#               Copyright (C) 2008-2009 The OpenAcoustics Team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from openacoustics.gmsh import *
from openacoustics.fem import *

def Demo2D(order = 1):
    # Gmsh geometry file name
    geo = 'square.geo'

    # Create an instance of class gmsh
    gmsh = GMsh2D()

    # Create an instance of class Fem
    fem = Fem2D()

    # Display the .geo file
    gmsh.viewGeo(geo)

    # Load the .geo file
    gmsh.loadGeo(geo, 2, order)


    # Load node indices and coordinates
    nodes = gmsh.getNodes()
    nodenum = nodes.shape[0] # Number of nodes

    # Get the source node number
    source = gmsh.getPoints("S")

    # Load node numbers of triangular elements
    elms = gmsh.getTriangles("", order)
    elmnum = elms.shape[0] # The number of triangular elements

    #gmsh.plotTriangularElements(elms,nodes[:,0],nodes[:,1])

    # Prepare global matrices for K and M
    K = sparse.lil_matrix((nodenum, nodenum))
    M = sparse.lil_matrix((nodenum, nodenum))

    # Start creating the global matrices
    # Append the triangular elements into the global matrices
    for N in range(elmnum):
        # Load the nodes that consists the Nth triangular element
        node1 = nodes[elms[N, 0], :]
        node2 = nodes[elms[N, 1], :]
        node3 = nodes[elms[N, 2], :]

        # Calculate the element matrices for the triangular element
        Kn, Mn, Bn, S = fem.triangularElement(node1, node2, node3, order)

        # Divide the mass matrix by the square of the sound speed
        # assuming air medium
        Mn = Mn / (343 * 343)

        idx = elms[N, :]
        for ii in range(Kn.shape[0]):
            for jj in range(Kn.shape[1]):
                K[idx[ii], idx[jj]] += Kn[ii, jj]
                M[idx[ii], idx[jj]] += Mn[ii, jj]

    # Frequency
    f = 170

    w = 2 * pi * f
    q = zeros(nodenum)

    # Unit divergence at the source location
    q[source[0]] = 1.0

    # Solve the global matrix system
    phi = linsolve.spsolve(K - (w**2) * M, q.T)

    # Extract x and y coordinates
    x = nodes[:, 0]
    y = nodes[:, 1]

    # Calculate the dB values
    #z = 20 * log10(abs(phi))

    # The linear absolute value
    z = real(phi)

    # Display spatial distribution of velocity potential with contour plot
    gmsh.contour(elms, x, y, z, "Relative sound pressure [dB]", "x[m]", "y[m]",
                 order)

def Demo3D(order=1):
    # Gmsh geometry file name
    geo = 'cube.geo'

    # Create an instance of class gmsh
    gmsh = GMsh3D()

    # Create an instance of class Fem
    fem = Fem3D()

    # Display the .geo file
    gmsh.viewGeo(geo)

    # Load the .geo file
    gmsh.loadGeo(geo, 0.8, order)

    # Load node indices and coordinates
    nodes = gmsh.getNodes()
    nodenum = nodes.shape[0] # Number of nodes

    # Get the source node number
    source = gmsh.getPoints("S")

    # Load node numbers of tetrahedral elements
    elms = gmsh.getTetrahedra("", order)
    elmnum = elms.shape[0] # Number of tetrahedral elements

    #gmsh.plotTetrahedoElements(sfs,nodes[:,0],nodes[:,1],nodes[:,2])

    # Prepare global matrices for K and M
    K = sparse.lil_matrix((nodenum, nodenum))
    M = sparse.lil_matrix((nodenum, nodenum))

    print nodenum
    print elmnum
    
    # Start creating the global matrices

    Vs = zeros(elmnum, 'float64')

    t = time()
    print "Setting field condition"

    # Append the tetrahedral elements into the global matrices
    for N in range(len(elms)):
        # Load the nodes that consists the Nth tetrahedral element
        node1 = nodes[elms[N, 0], :]
        node2 = nodes[elms[N, 1], :]
        node3 = nodes[elms[N, 2], :]
        node4 = nodes[elms[N, 3], :]

        # Calculate the element matrices for the tetrahedral element
        Kn, Mn, V = fem.tetrahedralElement(node1, node2, node3, node4, order)

        Vs[N] = V

        # Divide the mass matrix by the square of the sound speed
        # assuming air medium
        Mn = Mn / (343 * 343)

        idx = elms[N, :]
        for ii in range(Kn.shape[0]):
            for jj in range(Kn.shape[1]):
                K[idx[ii], idx[jj]] += Kn[ii, jj]
                M[idx[ii], idx[jj]] += Mn[ii, jj]

    print str("Setting field condition" + str(time() - t) + "sec")

    # Frequency
    f = 170

    w = 2 * pi * f
    q = zeros(nodenum)

    # Unit divergence at the source location
    q[source[0]] = -1.0

    # Solve the global matrix system
    t = time()
    phi, err = scipy.linalg.iterative.cg((K - (w**2) * M).tocsr(), q.T)
    print time() - t

    # Extract x, y and z coordinates
    x = nodes[:, 0]
    y = nodes[:, 1]
    z = nodes[:, 2]

    # Calculate the dB values
    scalars = 20 * log10(abs(phi))

    alpha = scalars
    alpha -= min(scalars)
    alpha /= max(scalars) - min(scalars)

    # Display spatial distribution of velocity potential with contour plot
    #gmsh.contour3(elms, x, y, z, scalars, "Relative sound pressure [dB]",
    #              "x[m]", "y[m]", order)

    gmsh.beginPostprocessing()
    gmsh.beginView("Slice of sound field")
    
    if order==1:
        gmsh.addTetrahedron(elms, x, y, z, scalars)
    elif order==2:
        gmsh.addTetrahedron2nd(elms, x, y, z, scalars)
    gmsh.closeView()
    gmsh.addCutPlaneByABCDForm(-1, 1, 1, 1, -1)
    gmsh.title(1, "Slice of potential velocity field")
    gmsh.xLabel(1, "x[m]")
    gmsh.yLabel(1, "y[m]")
    gmsh.zLabel(1, "z[m]")
    gmsh.addStyle()
    gmsh.merge(geo)
    gmsh.endPostprocessing()
    gmsh.run()

# Solve two dimensional rectangular room with first-order triangular elements
#Demo2D(1)

# Solve two dimensional rectangular room with second-order triangular elements
#Demo2D(2)

Demo3D(1)

