#  ooo  oooo  oooo o   o     o    ooo   ooo  o   o  oooo ooooo o  ooo   oooo
# 0   0 0   0 0    00  0    0 0  0   0 0   0 0   0 0       0   0 0   0 0
# 0   0 oooo  oooo 0 0 0   0   0 0     0   0 0   0  oooo   0   0 0      oooo
# 0   0 o     0    0 o 0   ooooo 0   o 0   0 0   0      0  0   0 0   o      0
#  ooo  o     oooo o  0o  0     0 ooo   ooo   ooo  0oooo   0   0  ooo  0oooo
#
#  (  (  (  (  (  (  (  (  (  (  (  (  O  )  )  )  )  )  )  )  )  )  )  )  )
#
#    OpenAcoustics: The Open Source Collection for Computational Acoustics
#
#               Copyright (C) 2008-2009 The OpenAcoustics Team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from numpy import array, dot, float64, linspace
from scipy import pi, sqrt, zeros, ones, random, mod, cross
from math import cos, sin, acos
from scipy.linalg import norm
from StringIO import StringIO
import csv
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot

#--------------------------------------------------------------------------------
#                       1. reading data fuction series
#--------------------------------------------------------------------------------
####################################
#### 1.1 read room data         ####
####################################
# @param 1 : file name of room data 
#
def readRoomData( roomDataFileName ):
    file = open(roomDataFileName, "r")

    ### read point data
    global numberOfPoints
    global point

    numberOfPoints = int( file.readline() )
    print "number of point: ", numberOfPoints
    point = zeros( (numberOfPoints,3), "float64")

    for m in range(numberOfPoints):
        itemList = file.readline().split(',')
        for n in range(len(itemList)):
            if(n == 0):
                num = int(itemList[n])
            else:
                point[m][n-1] = float64(itemList[n])
    
    ### read wall data
    global numberOfWalls
    global absorptionCoefficientIndex
    global numberOfApexes
    global wall
  
    numberOfWalls = int(file.readline())
    numberOfApexes = zeros(numberOfWalls, "int")
    wall = numberOfWalls*[0]
    absorptionCoefficientIndex = zeros(numberOfWalls, "int")

    for m in range(numberOfWalls):
        itemList = file.readline().split(',')
        wall[m] =  (len(itemList) - 3)*[0] 
        for n in range(len(itemList)):
            if(n == 0):
                wallIndex = int(itemList[n])
            elif(n == 1):
                absorptionCoefficientIndex[m] = int(itemList[n])
            elif(n == 2):
                numberOfApexes[m] = int(itemList[n])
            else:
                wall[m][n-3] = int(itemList[n])
    file.close()

###############################################
#### 1.2 read absorption coefficient data  ####
###############################################
# @param 1 :file name of absorption coefficient
#
def readAbsorptionCoefficientData( absorptionCoefficientDataFileName ):
    file = open(absorptionCoefficientDataFileName, "r")

    global numberOfMaterials
    global absorptionCoefficient

    numberOfMaterials = int(file.readline())

    absorptionCoefficient = zeros((numberOfMaterials,6), "float64")

    for m in range(numberOfMaterials):
        itemList = file.readline().split(',')
        for n in range(6):
            absorptionCoefficient[m][n] = float64(itemList[n+2])

    file.close()

##############################################
#### 1.3 read calculation condtion data   ####
##############################################
# @param 1: file name of calculation condition
#
def readCalculationConditionData( calculationConditionFileName ):
    file  = open(calculationConditionFileName, "r")

    global soundSpeed
    global soundSourcePosition
    global numberOfSoundRays
    global receivePosition
    global radiusReceivePoint
    global limitTimesOfReflection
    global calculationTotalTime
    global echoTimeStep
    
    soundSpeed = float(file.readline().split(',')[1])

    soundSourcePosition = zeros(3, "float64")
    itemList = file.readline().split(',')[1:4]
    for n in range(3):
        soundSourcePosition[n] = float64(itemList[n]);
    
    numberOfSoundRays = int(file.readline().split(',')[1])
    
    receivePosition = zeros(3, "float64")
    itemList = file.readline().split(',')[1:4]
    for n in range(3):
        receivePosition[n] = float64(itemList[n]);

    radiusReceivePoint = float64(file.readline().split(',')[1])

    limitTimesOfReflection = int(file.readline().split(',')[1])
                               
    calculationTotalTime = float64(file.readline().split(',')[1])

    echoTimeStep = float64(file.readline().split(',')[1])

#--------------------------------------------------------------------------------
#2. geometrical function series for ray tracing
#--------------------------------------------------------------------------------
##########################################
##### 2.1 vectorization from 2 points ####
##########################################
# @param p1: point 1 (x1, y1, z1)
# @param p2: point 2 (x2, y2, z2)
#
def pointToVector( p1, p2 ):
    v = array(p2 - p1)
    return v

######################################
#### 2.2 calculate plane equation ####
######################################
#
def calcPlaneEquation():
    global planeEquation
    planeEquation = zeros((numberOfWalls,4), "float64")
    for m in range(numberOfWalls):
        p1 = point[wall[m][0]]
        p2 = point[wall[m][1]]
        p3 = point[wall[m][2]]

        v1 = pointToVector(p1, p2)
        v2 = pointToVector(p3, p2)

        normalVector = cross(v1, v2)
        normalVector /= norm(normalVector)
        planeEquation[m][0:3] = normalVector
        d = (-1.0) * (normalVector[0]*p1[0] + normalVector[1]*p1[1] + normalVector[2]*p1[2])
        planeEquation[m][3] = d
        print planeEquation[m]

###############################################################
#### 2.3 calculate cross point from one plane and one line ####
###############################################################
# @param 1 v: vector of sound ray (vx, vy, vz)
# @param 2 p: point on sound ray (x1, y1, z1)
# @param 3 pe: plane equation coefficient (a, b, c, d)
#
def calcCrossPoint(v, p, pe):
    crossPoint = zeros(3, "float64")
    a, b, c, d = pe
    t = -(a*p[0] + b*p[1] + c*p[2] + d) / (a*v[0] + b*v[1] + c*v[2])
    for n in range(3):
        crossPoint[n] = p[n] + t*v[n]
    return crossPoint

################################################
#### 2.4 calculate the distance of 2 points ####
################################################
# @param p1: point1 (x1, y1, z1)
# @param p2: point2 (x2, y2, z2)
# 
def calcPointToPointDistance(p1, p2):
    d = norm(p2 - p1)
    return d

###########################################################
#### 2.5 calculate the distance from a point to a line ####
###########################################################
# @param v: the vector of a line (vx, vy, vz)
# @param p1: a point on a line (x1, y1, z1)
# @param p2: a point away from a line (x2, y2, z2)
#
def calcPointToLineDistance(v, p1, p2): 
    t =  dot(v,(p2-p1)) / dot(v,v)
    h = t * v + p1                          ## cross point: h(x, y, z)
    d = calcPointToPointDistance(h, p2)
    return d

#################################
#### 2.6 create sound source  ###
#################################
#
def createSoundSource():
    global rayVector
    global rayPosition
    global actualNumberOfSoundRays
    rayVector = zeros((numberOfSoundRays,3), "float64")
    rayPosition = zeros((numberOfSoundRays,3), "float64")
    dOmega = 4.0*pi/numberOfSoundRays
    dPhi = sqrt(2.0*dOmega/sqrt(3.0))
    dTheta = sqrt(3.0)*dPhi/2.0
    theta = -pi/2.0
    actualNumberOfSoundRays = 0
    counter = 0;
    while(theta <= pi/2.0):
        numberOfRaysOnHorizontalPlane = int(round(cos(theta)*2.0*pi/dPhi))
        if(actualNumberOfSoundRays + numberOfRaysOnHorizontalPlane <= numberOfSoundRays):
            for n in range(actualNumberOfSoundRays, 
                           actualNumberOfSoundRays + numberOfRaysOnHorizontalPlane):
                if(mod(counter,2) == 0):
                    rayVector[n][0] = cos(2*pi*n/numberOfRaysOnHorizontalPlane)*cos(theta)
                    rayVector[n][1] = sin(2*pi*n/numberOfRaysOnHorizontalPlane)*cos(theta)
                    rayVector[n][2] = sin(theta)
                else:
                    rayVector[n][0] = cos(2*pi*(n+1/2)/numberOfRaysOnHorizontalPlane)*cos(theta)
                    rayVector[n][1] = sin(2*pi*(n+1/2)/numberOfRaysOnHorizontalPlane)*cos(theta)
                    rayVector[n][2] = sin(theta)
                rayPosition[n] = soundSourcePosition
            actualNumberOfSoundRays += numberOfRaysOnHorizontalPlane
        if(actualNumberOfSoundRays == numberOfSoundRays):
            break
        theta += dTheta
        counter += 1
    print 'Actual number of Sound Rays: ', actualNumberOfSoundRays

#-------------------------------------------------------------------------------------
#                   3. collision detection function series
#-------------------------------------------------------------------------------------
############################################################
### 3.1 Is a sound ray going toward or backward a wall? ####
############################################################
# @param v1: vector of a sound ray
# @param v2: normal vector of a wall
#   
def isGoingTowardWall(v1, v2):
    if(dot(v1, v2) < 0.0):
        return True # toward
    else:
        return False # backward

######################################
### 3.2 Is a point inside a wall? ####
######################################
# @param p1: a point on a plane
# @param wallIndex: Index of a wall
#
def isInsideWall(p1, wallIndex):
    totalDegreeOfAngles = 0.0
    for m in range(numberOfApexes[wallIndex] - 1):
        v1 = pointToVector(point[wall[wallIndex][m]], p1)
        v2 = pointToVector(point[wall[wallIndex][m+1]], p1)
        angle = acos( dot(v1,v2) / (norm(v1)*norm(v2)) ) 
        isClockwise = dot( cross(v1,v2), planeEquation[wallIndex][0:3] )
        if( isClockwise > 0.0):
               angle *= -1
        totalDegreeOfAngles += angle
    v1 = pointToVector(point[wall[wallIndex][numberOfApexes[wallIndex] - 1]], p1)
    v2 = pointToVector(point[wall[wallIndex][0]], p1)
    angle = acos( dot(v1,v2) / (norm(v1)*norm(v2)) )
    isClockwise = dot( cross(v1,v2), planeEquation[wallIndex][0:3] )
    if( isClockwise > 0.0):
        angle *= -1
    totalDegreeOfAngles += angle
      
    if(totalDegreeOfAngles > 2.0*pi*0.99 and \
       totalDegreeOfAngles < 2.0*pi*1.01):    
                            # 0.99 and 1.01 is magic number
        return True
    else:
        return False

#########################################
### 3.3 calculate reflection vector #####
#########################################
# @param vi: incident vector of line
# @param vn: normal vector of plane
#
def calcReflectionVector(vi, vn):
    reflectionVector = vi - 2 * dot(vn,vi) * vn
    reflectionVector  /= norm(reflectionVector)
    return reflectionVector

#---------------------------------------------------------------------------
#                               4. main routine
#---------------------------------------------------------------------------
print "#1 Read Room Data"
readRoomData("room.csv")

print "#2 Read Absorption Coefficient Data"
readAbsorptionCoefficientData("absorption.csv")

print "#3 Read Calculation Condition"
readCalculationConditionData("control4sample1.csv")

print "#4 Calculate Plane Equation"
calcPlaneEquation()

print "#5 Create Sound Souce"
createSoundSource()

print "#6 Shoot off Sound Rays!!!!!!"

#######################################
### define local varies
#######################################
## propageted distance
distance = zeros((actualNumberOfSoundRays,1), "float64")

## number of reflection times
reflectionTimes = zeros((actualNumberOfSoundRays, 1), "int")

## maximum distance for calculation
limitDistance = calculationTotalTime * soundSpeed

# energy of all sound particle in each 1/1 octave band (125 to 4k Hz)
energy = ones((actualNumberOfSoundRays,6), "float64") / actualNumberOfSoundRays

## closs point coordinates
crossPoint = zeros((actualNumberOfSoundRays,3), "float64")

## temporary cross point coordinates
tmpCrossPoint = zeros((actualNumberOfSoundRays,3), "float64")

## minimum length to cross point
minLength = zeros((actualNumberOfSoundRays,1), "float64")

## index of reflection wall
crossWallIndex = zeros((actualNumberOfSoundRays,1), "int")

## closs point coordinates for every reflection time 
crossPointArchive = zeros((actualNumberOfSoundRays, limitTimesOfReflection+1, 3), "float64")
for n in range(actualNumberOfSoundRays):
    crossPointArchive[n][0] = soundSourcePosition

### number of elements of array in echo time pattern  
numberOfElementsOfEchoTime = int(calculationTotalTime/echoTimeStep+1)

### echo time pattern in each 1/1 octave band (125 to 4k Hz)
echoTime = zeros((numberOfElementsOfEchoTime, 6), "float64")

###################################################
#   main loop
###################################################
for m in range(actualNumberOfSoundRays):
    while( distance[m] < limitDistance and reflectionTimes[m] < limitTimesOfReflection):
        #-------------------------------
        #  search crossing wall
        #-------------------------------
        for n in range(numberOfWalls):
            if( isGoingTowardWall( rayVector[m], planeEquation[n][0:3] ) == True ):
                tmpCrossPoint = calcCrossPoint(rayVector[m], rayPosition[m], planeEquation[n])
                if( norm(tmpCrossPoint) < 10.0**3.0 ):
                    if( isInsideWall( tmpCrossPoint, n ) == True ):
                        tmpLength = calcPointToPointDistance(tmpCrossPoint, rayPosition[m])
                        if(minLength[m] == 0 or minLength[m] > tmpLength):
                            minLength[m] = tmpLength
                            crossWallIndex[m] = n
                            crossPoint[m] = tmpCrossPoint

        #----------------------------------------------------
        #  Does a sound ray cross the receiving sphere?
        #----------------------------------------------------
        v1 = pointToVector(rayPosition[m], receivePosition)
        d = calcPointToLineDistance(rayVector[m], rayPosition[m], receivePosition)
        if(dot(v1, rayVector[m]) > 0 and d <= radiusReceivePoint):
            time = (distance[m] + calcPointToPointDistance(rayPosition[m], receivePosition)) / soundSpeed
            boxIndex = int(time / echoTimeStep)
            if( boxIndex < numberOfElementsOfEchoTime ):
                echoTime[boxIndex,:] += energy[m,:]
                
        #------------------------------------------------
        # reflect and refine sound ray properties     
        #------------------------------------------------
        if(minLength[m] != 0):
            rayVector[m] = calcReflectionVector(rayVector[m], planeEquation[int(crossWallIndex[m])][0:3])
            distance[m] += minLength[m]
            reflectionTimes[m] += 1
            rayPosition[m] = crossPoint[m]
            energy[m,:] *= (1 - absorptionCoefficient[absorptionCoefficientIndex[int(crossWallIndex[m])]])
            crossPointArchive[m, reflectionTimes[m], :]  = crossPoint[m]
            minLength[m] = 0.0
        else:
            print "error!!"
            exit()

#---------------------------------------------
#    drawing results
#---------------------------------------------
### drawing ray tracing results
fig1 = matplotlib.pyplot.figure(1)
ax1 = Axes3D(fig1)
matplotlib.pyplot.axis('off')

for m in range(actualNumberOfSoundRays):
    x = crossPointArchive[m,:,0]
    y = crossPointArchive[m,:,1]
    z = crossPointArchive[m,:,2]
    ax1.plot(x,y,z,color=[float(actualNumberOfSoundRays-m)/actualNumberOfSoundRays,
                          float(abs(actualNumberOfSoundRays/2-m))/actualNumberOfSoundRays,
                          float(m)/actualNumberOfSoundRays])

ax1.set_xlim3d([min(point[:,0]), max(point[:,0])])
ax1.set_ylim3d([min(point[:,0]), max(point[:,0])])
ax1.set_zlim3d([min(point[:,0]), max(point[:,0])])

fig1.show()
matplotlib.pyplot.draw()
matplotlib.pyplot.axis('equal')



### drawing impulse responses
fig2 = matplotlib.pyplot.figure(2)
timeScale = linspace(0, calculationTotalTime, numberOfElementsOfEchoTime)
legend = ['125 Hz', '250 Hz', '500 Hz', '1 kHz', '2 kHz', '4 kHz']
for n in range(6):
    matplotlib.pyplot.subplot('32'+str(n+1))
    matplotlib.pyplot.bar(timeScale, echoTime[:,n], width=echoTimeStep*0.9)
    matplotlib.pyplot.xlim([0, calculationTotalTime])
    matplotlib.pyplot.ylim([0, echoTime.max()])
    matplotlib.pyplot.text(0.8*calculationTotalTime, 0.8*echoTime.max(), legend[n])
    matplotlib.pyplot.ylabel('energy')
    if n==4 or n==5:
        matplotlib.pyplot.xlabel('time (s)')

fig2.show()
