#  ooo  oooo  oooo o   o     o    ooo   ooo  o   o  oooo ooooo o  ooo   oooo
# 0   0 0   0 0    00  0    0 0  0   0 0   0 0   0 0       0   0 0   0 0
# 0   0 oooo  oooo 0 0 0   0   0 0     0   0 0   0  oooo   0   0 0      oooo
# 0   0 o     0    0 o 0   ooooo 0   o 0   0 0   0      0  0   0 0   o      0
#  ooo  o     oooo o  0o  0     0 ooo   ooo   ooo  0oooo   0   0  ooo  0oooo
#
#  (  (  (  (  (  (  (  (  (  (  (  (  O  )  )  )  )  )  )  )  )  )  )  )  )
#
#    OpenAcoustics: The Open Source Collection for Computational Acoustics
#
#               Copyright (C) 2008-2012 The OpenAcoustics Team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from numpy import array, dot, float64, linspace
from scipy import pi, sqrt, zeros, ones, random, mod, cross
from math import cos, sin, acos
from scipy.linalg import norm
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot

#--------------------------------------------------------------------------------
#                       1. reading data fuction series
#--------------------------------------------------------------------------------
####################################
#### 1.1 read room data         ####
####################################
# @param 1 : file name of room data 
#
def readRoomData( roomDataFileName ):
    file = open(roomDataFileName, "r")

    ### read point data
    global numberOfPoints
    global point

    numberOfPoints = int( file.readline() )
    print "number of point: ", numberOfPoints
    point = zeros( (numberOfPoints,3), "float64")

    for m in range(numberOfPoints):
        itemList = file.readline().split(',')
        for n in range(len(itemList)):
            if(n == 0):
                num = int(itemList[n])
            else:
                point[m][n-1] = float64(itemList[n])
    
    ### read wall data
    global numberOfWalls
    global absorpCoeffIndex
    global numberOfApexes
    global wall
  
    numberOfWalls = int(file.readline())
    numberOfApexes = zeros(numberOfWalls, "int")
    wall = numberOfWalls*[0]
    absorpCoeffIndex = zeros(numberOfWalls, "int")

    for m in range(numberOfWalls):
        itemList = file.readline().split(',')
        wall[m] =  (len(itemList) - 3)*[0] 
        for n in range(len(itemList)):
            if(n == 0):
                wallIndex = int(itemList[n])
            elif(n == 1):
                absorpCoeffIndex[m] = int(itemList[n])
            elif(n == 2):
                numberOfApexes[m] = int(itemList[n])
            else:
                wall[m][n-3] = int(itemList[n])
    file.close()

###############################################
#### 1.2 read absorption coefficient data  ####
###############################################
# @param 1 :file name of absorption coefficient
#
def readAbsorpCoeffData( absorpCoeffDataFileName ):
    file = open(absorpCoeffDataFileName, "r")

    global numberOfMaterials
    global absorpCoeff

    numberOfMaterials = int(file.readline())

    absorpCoeff = zeros((numberOfMaterials,6), "float64")

    for m in range(numberOfMaterials):
        itemList = file.readline().split(',')
        for n in range(6):
            absorpCoeff[m][n] = float64(itemList[n+2])

    file.close()

##############################################
#### 1.3 read calculation condtion data   ####
##############################################
# @param 1: file name of calculation condition
#
def readCalcCondData( calcCondFileName ):
    file  = open(calcCondFileName, "r")

    global soundSpeed
    global soundSourcePosition
    global numberOfRays
    global receiverPosition
    global radiusReceivingPoint
    global limitNumOfReflectTimes
    global totalTimeOfCalc
    global echoTimeInterval
    
    soundSpeed = float(file.readline().split(',')[1])

    soundSourcePosition = zeros(3, "float64")
    itemList = file.readline().split(',')[1:4]
    for n in range(3):
        soundSourcePosition[n] = float64(itemList[n]);
    numberOfRays = int(file.readline().split(',')[1])
    
    receiverPosition = zeros(3, "float64")
    itemList = file.readline().split(',')[1:4]
    for n in range(3):
        receiverPosition[n] = float64(itemList[n]);

    radiusReceivingPoint = float64(file.readline().split(',')[1])

    limitNumOfReflectTimes = int(file.readline().split(',')[1])
                               
    totalTimeOfCalc = float64(file.readline().split(',')[1])

    echoTimeInterval = float64(file.readline().split(',')[1])


#--------------------------------------------------------------------------------
#2. geometrical function series for ray tracing
#--------------------------------------------------------------------------------
##########################################
##### 2.1 vectorization from 2 points ####
##########################################
# @param p1: point 1 (x1, y1, z1)
# @param p2: point 2 (x2, y2, z2)
#
def pointToVector( p1, p2 ):
    v = array(p2 - p1)
    return v

######################################
#### 2.2 calculate plane equation ####
######################################
#
def calcPlaneEq():
    global planeEq
    planeEq = zeros((numberOfWalls,4), "float64")
    for m in range(numberOfWalls):
        p1 = point[wall[m][0]]
        p2 = point[wall[m][1]]
        p3 = point[wall[m][2]]

        v1 = pointToVector(p2, p1)
        v2 = pointToVector(p2, p3)

        normalVector = cross(v1, v2)
        normalVector /= norm(normalVector)
        planeEq[m][0:3] = normalVector
        d = (-1.0) * (normalVector[0]*p1[0] 
                    + normalVector[1]*p1[1] 
                    + normalVector[2]*p1[2])
        planeEq[m][3] = d
        print planeEq[m]

###############################################################
#### 2.3 calculate cross point from one plane and one line ####
###############################################################
# @param 1 v: vector of sound ray (vx, vy, vz)
# @param 2 p: point on sound ray (x1, y1, z1)
# @param 3 pe: plane equation coefficient (a, b, c, d)
#
def calcCrossPoint(v, p, pe):
    crossPoint = zeros(3, "float64")
    a, b, c, d = pe
    t = -(a*p[0] + b*p[1] + c*p[2] + d) / (a*v[0] + b*v[1] + c*v[2])
    for n in range(3):
        crossPoint[n] = p[n] + t*v[n]
    return crossPoint

##################################################
#### 2.4 calculate the distance of two points ####
##################################################
# @param p1: point1 (x1, y1, z1)
# @param p2: point2 (x2, y2, z2)
# 
def calcPointToPointDistance(p1, p2):
    d = norm(p2 - p1)
    return d

###########################################################
#### 2.5 calculate the distance from a point to a line ####
###########################################################
# @param v: the vector of a line (vx, vy, vz)
# @param p1: a point on a line (x1, y1, z1)
# @param p2: a point away from a line (x2, y2, z2)
#
def calcPointToLineDistance(v, p1, p2): 
    t =  dot(v,(p2-p1)) / dot(v,v)
    h = t * v + p1                          ## cross point: h(x, y, z)
    d = calcPointToPointDistance(h, p2)
    return d

#################################
#### 2.6 create sound source  ###
#################################
#
def createOmniSoundSource():
    global rayVector
    global rayPosition
    global actualNumberOfRays
    rayVector = zeros((numberOfRays,3), "float64")
    rayPosition = zeros((numberOfRays,3), "float64")
    d_omega = 4.0*pi/numberOfRays
    d_phi = sqrt(2.0*d_omega/sqrt(3.0))
    d_theta = sqrt(3.0)*d_phi/2.0
    theta = -pi/2.0
    actualNumberOfRays = 0
    counter = 0;
    while(theta <= pi/2.0):
        numberOfRaysOn_H_plane = int(round(cos(theta)*2.0*pi/d_phi))
        if(actualNumberOfRays + numberOfRaysOn_H_plane <= numberOfRays):
            for n in range(actualNumberOfRays, 
                           actualNumberOfRays + numberOfRaysOn_H_plane):
                if(mod(counter,2) == 0):
                    m = n
                else:
                    m = n + 1/2
                rayVector[n][0] = cos(2*pi*m/numberOfRaysOn_H_plane)*cos(theta)
                rayVector[n][1] = sin(2*pi*m/numberOfRaysOn_H_plane)*cos(theta)
                rayVector[n][2] = sin(theta)
                rayPosition[n] = soundSourcePosition
            actualNumberOfRays += numberOfRaysOn_H_plane
        if(actualNumberOfRays == numberOfRays):
            break
        theta += d_theta
        counter += 1
    print 'Actual number of Sound Rays: ', actualNumberOfRays
    return actualNumberOfRays


#-------------------------------------------------------------------------------------
#                   3. collision detection function series
#-------------------------------------------------------------------------------------
######################################################
### 3.1 Is a sound ray toward or backward a wall? ####
######################################################
# @param v1: vector of a sound ray
# @param v2: normal vector of a wall
#   
def isTowardWall(v1, v2):
    if(dot(v1, v2) < 0.0):
        return True # toward
    else:
        return False # backward

######################################
### 3.2 Is a point inside a wall? ####
######################################
# @param p1: a point on a plane
# @param wallIndex: Index of a wall
#
def isInsideWall(p1, wallIndex):
    totalDegreeOfAngles = 0.0
    for m in range(numberOfApexes[wallIndex] - 1):
        v1 = pointToVector(point[wall[wallIndex][m]], p1)
        v2 = pointToVector(point[wall[wallIndex][m+1]], p1)
        angle = acos( dot(v1,v2) / (norm(v1)*norm(v2)) ) 
        isClockwise = dot( cross(v1,v2), planeEq[wallIndex][0:3] )
        if( isClockwise > 0.0 ):
               angle *= -1
        totalDegreeOfAngles += angle
    v1 = pointToVector(point[wall[wallIndex][numberOfApexes[wallIndex] - 1]], p1)
    v2 = pointToVector(point[wall[wallIndex][0]], p1)
    angle = acos( dot(v1,v2) / (norm(v1)*norm(v2)) )
    isClockwise = dot( cross(v1,v2), planeEq[wallIndex][0:3] )
    if( isClockwise > 0.0 ):
        angle *= -1
    totalDegreeOfAngles += angle
      
    if(totalDegreeOfAngles > 2.0*pi*0.99 and \
       totalDegreeOfAngles < 2.0*pi*1.01):    
                            # 0.99 and 1.01 is magic number
        return True
    else:
        return False

#########################################
### 3.3 calculate reflection vector #####
#########################################
# @param vi: incident vector of line
# @param vn: normal vector of plane
#
def calcReflectionVector(vi, vn):
    reflectionVector = vi - 2 * dot(vn,vi) * vn
    reflectionVector  /= norm(reflectionVector)
    return reflectionVector



#--------------------------------------------------------
#               4. main routine
#--------------------------------------------------------
print "#1 Read Room Data"
readRoomData("room.csv")

print "#2 Read Absorption Coefficient Data"
readAbsorpCoeffData("absorption.csv")

print "#3 Read Condition for Calculation"
readCalcCondData("control4sample1.csv")

print "#4 Calculate coefficients of Plane Equation"
calcPlaneEq()

print "#5 Create Sound Souce"
numberOfRays = createOmniSoundSource()

print "#6 Shoot off Sound Rays!!!!!!"

#######################################
### define local varies
#######################################
## propageted distance
distance = zeros((numberOfRays,1), "float64")

## number of reflection times
reflectionTimes = zeros((numberOfRays, 1), "int")

## maximum distance for calculation
limitDistance = totalTimeOfCalc * soundSpeed

# energy of all sound particle in each 1/1 octave band (125 to 4k Hz)
energy = ones((numberOfRays,6), "float64") / numberOfRays

## closs point coordinates
crossPoint = zeros((numberOfRays,3), "float64")

## temporary cross point coordinates
tmpCrossPoint = zeros((numberOfRays,3), "float64")

## minimum length to cross point
minDistance = 0.0

## index of reflection wall
crossWallIndex = zeros((numberOfRays,1), "int")

## closs point coordinates for every reflection time 
crossPointArchive = zeros((numberOfRays, limitNumOfReflectTimes+1, 3), "float64")
for n in range(numberOfRays):
    crossPointArchive[n][0] = soundSourcePosition

### number of elements of array in echo time pattern  
numberOfElemofEchoTime = int(totalTimeOfCalc/echoTimeInterval+1)

### echo time pattern in each 1/1 octave band (125 to 4k Hz)
echoTime = zeros((numberOfElemofEchoTime, 6), "float64")

###################################################
#   main loop
###################################################
for m in range(numberOfRays):
    minDistance = 0.0
    while( distance[m] < limitDistance 
           and reflectionTimes[m] < limitNumOfReflectTimes):
        #-------------------------------
        #  search crossing wall
        #-------------------------------
        for n in range(numberOfWalls):
            if( isTowardWall( rayVector[m], planeEq[n][0:3] ) == True ):
                tmpCrossPoint = \
                    calcCrossPoint(rayVector[m], rayPosition[m], planeEq[n])
                if( norm(tmpCrossPoint) < 10.0**3.0 ):
                                                ### avoid math domain error
                                                ### 10.0 ** 3.0 is magick number
                    if( isInsideWall( tmpCrossPoint, n ) == True ):
                        tmpDistance = \
                            calcPointToPointDistance(tmpCrossPoint, rayPosition[m])
                        if(minDistance == 0.0 or minDistance > tmpDistance):
                            minDistance = tmpDistance
                            crossWallIndex[m] = n
                            crossPoint[m] = tmpCrossPoint

        #----------------------------------------------------
        #  Does a sound ray cross the receiving sphere?
        #----------------------------------------------------
        v1 = pointToVector(rayPosition[m], receiverPosition)
        d = calcPointToLineDistance(rayVector[m], rayPosition[m], receiverPosition)
        if(dot(v1, rayVector[m]) > 0 and d <= radiusReceivingPoint):
            time = (distance[m] \
                    + calcPointToPointDistance(rayPosition[m], receiverPosition)) \
                    / soundSpeed
            boxIndex = int(time / echoTimeInterval)
            if( boxIndex < numberOfElemofEchoTime ):
                echoTime[boxIndex,:] += energy[m,:]
                
        #------------------------------------------------
        # reflect and refine sound ray properties     
        #------------------------------------------------
        if(minDistance != 0):
            rayVector[m] = \
              calcReflectionVector(rayVector[m], planeEq[int(crossWallIndex[m])][0:3])
            distance[m] += minDistance
            reflectionTimes[m] += 1
            rayPosition[m] = crossPoint[m]
            energy[m,:] *= (1 - absorpCoeff[absorpCoeffIndex[int(crossWallIndex[m])]])
            crossPointArchive[m, reflectionTimes[m], :]  = crossPoint[m]
            minDistance = 0.0
        else:
            print "error!!"
            exit()

#---------------------------------------------
#    drawing results
#---------------------------------------------
### drawing ray tracing results
#fig1 = matplotlib.pyplot.figure(1, figsize=(5,5))
#ax1 = Axes3D(fig1)
#
#for m in range(numberOfRays):
#    x = crossPointArchive[m,:,0]
#    y = crossPointArchive[m,:,1]
#    z = crossPointArchive[m,:,2]
#    ax1.plot(x,y,z, color=[float(numberOfRays-m)/numberOfRays,
#                           float(abs(numberOfRays/2-m))/numberOfRays,
#                           float(m)/numberOfRays])
#matplotlib.pyplot.draw()

### drawing impulse responses
fig2 = matplotlib.pyplot.figure(2)
timeScale = linspace(0, totalTimeOfCalc, numberOfElemofEchoTime)
legend = ['125 Hz', '250 Hz', '500 Hz', '1 kHz', '2 kHz', '4 kHz']
for n in range(6):
    matplotlib.pyplot.subplot('32'+str(n+1))
    matplotlib.pyplot.bar(timeScale, echoTime[:,n], width=echoTimeInterval*0.9)
    matplotlib.pyplot.xlim([0, totalTimeOfCalc])
    matplotlib.pyplot.ylim([0, echoTime.max()])
    matplotlib.pyplot.text(0.8*totalTimeOfCalc, 0.8*echoTime.max(), legend[n])
    matplotlib.pyplot.ylabel('energy')
    if n==4 or n==5:
        matplotlib.pyplot.xlabel('time (s)')

matplotlib.pyplot.show()

