#  ooo  oooo  oooo o   o     o    ooo   ooo  o   o  oooo ooooo o  ooo   oooo
# 0   0 0   0 0    00  0    0 0  0   0 0   0 0   0 0       0   0 0   0 0
# 0   0 oooo  oooo 0 0 0   0   0 0     0   0 0   0  oooo   0   0 0      oooo
# 0   0 o     0    0 o 0   ooooo 0   o 0   0 0   0      0  0   0 0   o      0
#  ooo  o     oooo o  0o  0     0 ooo   ooo   ooo  0oooo   0   0  ooo  0oooo
#
#  (  (  (  (  (  (  (  (  (  (  (  (  O  )  )  )  )  )  )  )  )  )  )  )  )
#
#    OpenAcoustics: The Open Source Collection for Computational Acoustics
#
#               Copyright (C) 2008-2009 The OpenAcoustics Team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

## @package openacoustics.fem
# 2, 3 dimensional fem module for acoustic problem

__author__ = "hisaha <hisaha@souldsp.jp>"
__date__ = "2008/02/07"
__credits__ = "Hisharu Suzuki"
__version__ = "$Revision: 0 $"

from scipy import *
import scipy.linalg.iterative
from pylab import *
import math
import os

## for calculating 2-dimensional sound field with fem
class Fem2D:
    """A simple 2-dimensional fem class for calculating sound field with triangle element"""

    ##simple function for plot triangle element
    #  @param p1 cordinate vector of first node of element
    #  @param p2 cordinate vector of second node of element
    #  @param p3 cordinate vector of third node of element
    def elmplot(self, p1, p2, p3):
        """simple function for plot triangle element"""
        plot([p1[0], p2[0]], [p1[1], p2[1]])
        plot([p1[0], p3[0]], [p1[1], p3[1]])
        plot([p2[0], p3[0]], [p2[1], p3[1]])

    ##simple function for plot triangle element
    #  @param n1 cordinate vector of first node of element
    #  @param n2 cordinate vector of second node of element
    #  @param n3 cordinate vector of third node of element
    #  @param order order of element
    #  @retval K K matrix
    #  @retval M M matrix
    #  @retval B B matrix
    #  @retval S surface area
    def triangularElement(self, n1, n2, n3, order = 1):
        """calculate K,M matrix, B matrix(for transform cordinate), surface area"""
        if order == 2:
            return self.triangularElement2nd(n1, n2, n3)

        node = [n1, n2, n3]
        a = [0, 0, 0]
        b = [0, 0, 0]
        c = [0, 0, 0]

        for kk in range(3):
            k = mod(kk, 3)
            l = mod(kk + 1, 3)
            m = mod(kk + 2, 3)
            # Intermediate variables to calculate a, b, c and d
            xl = node[l][0]
            xm = node[m][0]
            yl = node[l][1]
            ym = node[m][1]
            ak = xl * ym - xm * yl
            bk = yl - ym
            ck = xm - xl
            a[k] = ak
            b[k] = bk
            c[k] = ck

        S = sum(a) / 2
        T1 = -1 / (2 * S) * (b[1] * b[2] + c[1] * c[2])
        T2 = -1 / (2 * S) * (b[0] * b[2] + c[0] * c[2])
        T3 = -1 / (2 * S) * (b[1] * b[0] + c[1] * c[0])

        K = array([[T2 + T3, -T3, -T2],
                   [-T3, T1 + T3, -T1],
                   [-T2, -T1, T1 + T2]]) / 2.

        M = array([[2., 1., 1.],
                   [1., 2., 1.],
                   [1., 1., 2.]]) * S / 12.

        B = array([[a[0], a[1], a[2]],
                   [b[0], b[1], b[2]],
                   [c[0], c[1], c[2]]])

        return K, M, B, S

    ## calculate 2nd order element's K,M matrix, B matrix(for transform cordinate), surface area
    #  @param n1 cordinate vector of first node of element
    #  @param n2 cordinate vector of second node of element
    #  @param n3 cordinate vector of third node of element
    #  @param order order of element
    #  @retval K K matrix
    #  @retval M M matrix
    #  @retval B B matrix
    #  @retval S surface area
    def triangularElement2nd(self, n1, n2, n3):
        """calculate 2nd order element's K,M matrix, B matrix(for transform cordinate), surface area"""
        node = [n1, n2, n3]
        a = [0, 0, 0]
        b = [0, 0, 0]
        c = [0, 0, 0]

        for kk in range(3):
            k = mod(kk, 3)
            l = mod(kk + 1, 3)
            m = mod(kk + 2, 3)
            # Intermediate variables to calculate a, b, c and d
            xl = node[l][0]
            xm = node[m][0]
            yl = node[l][1]
            ym = node[m][1]
            ak = xl * ym - xm * yl
            bk = yl - ym
            ck = xm - xl
            a[k] = ak
            b[k] = bk
            c[k] = ck

        S = sum(a) / 2
        T1 = -1 / (2 * S) * (b[1] * b[2] + c[1] * c[2])
        T2 = -1 / (2 * S) * (b[0] * b[2] + c[0] * c[2])
        T3 = -1 / (2 * S) * (b[1] * b[0] + c[1] * c[0])

#1,2,3,4,5,6
##        K = array([[3*(T2+T3),        T3,        T2,            0,        -4*T2,        -4*T3],
##                    [      T3, 3*(T1+T3),        T1,        -4*T1,            0,        -4*T3],
##                    [      T2,        T1, 3*(T1+T2),        -4*T1,        -4*T2,            0],
##                    [       0,     -4*T1,     -4*T1, 8*(T1+T2+T3),        -8*T3,        -8*T2],
##                    [   -4*T2,         0,     -4*T2,        -8*T3, 8*(T1+T2+T3),        -8*T1],
##                    [   -4*T3,     -4*T3,         0,        -8*T2,        -8*T1, 8*(T1+T2+T3)]])/6

        #1,2,3,6,4,5 (Toggle matrix for gmsh format)
        K = array([[3 * (T2 + T3), T3, T2, -4 * T3, 0, -4 * T2],
                   [T3, 3 * (T1 + T3), T1, -4 * T3, -4 * T1, 0],
                   [T2, T1, 3 * (T1 + T2), 0, -4 * T1, -4 * T2],
                   [-4 * T3, -4 * T3, 0, 8 * (T1 + T2 + T3), -8 * T2, -8 * T1],
                   [0, -4 * T1, -4 * T1, -8 * T2, 8 * (T1 + T2 + T3), -8 * T3],
                   [-4 * T2, 0, -4 * T2, -8 * T1, -8 * T3, 8 * (T1 + T2 + T3)]]
                  ) / 6

#1,2,3,4,5,6
##        M = array([[6.,-1.,-1.,-4., 0., 0.],
##                    [-1., 6.,-1., 0.,-4., 0.],
##                    [-1.,-1., 6., 0., 0.,-4.],
##                    [-4., 0., 0.,32.,16.,16.],
##                    [0.,-4., 0.,16.,32.,16.],
##                    [0., 0.,-4.,16.,16.,32.]])*S/180.

        #1,2,3,6,4,5 (Toggle matrix for gmsh format)
        M = array([[6., -1., -1., 0., -4., 0.],
                   [-1., 6., -1., 0., 0., -4.],
                   [-1., -1., 6., -4., 0., 0.],
                   [0., 0., -4., 32., 16., 16.],
                   [-4., 0., 0., 16., 32., 16.],
                   [0., -4., 0., 16., 16., 32.]]) * S / 180.

        B = array([[a[0], a[1], a[2]],
                   [b[0], b[1], b[2]],
                   [c[0], c[1], c[2]]])

        return K, M, B, S

    ## for calculating quantities at arbitrary point.
    # @param abpoint cordinate of arbitrary point
    # @param potential quantities of nodes
    # @elms index matrix of element
    # @elms cordinate matrix of element's nodes
    # @retval phi potential quantity at arbitrary point
    # @retval phix partial difference of potential quantity in x-direction at arbitrary point
    # @retval phiy partial difference of potential quantity in y-direction at arbitrary point
    def getAtArbitraryPoint(self, abpoint, phis, elms, nodes):
        """for calculating quantities at arbitrary point."""
        elmnum = elms.shape[0]

        e = -1

        for N in range(elmnum):
            # Load the nodes that consist the Nth element
            node1 = nodes[elms[N, 0], 0:2]
            node2 = nodes[elms[N, 1], 0:2]
            node3 = nodes[elms[N, 2], 0:2]

            if self.isInner(abpoint, node1, node2, node3):
                e = N
                break

        if e < 0:
            return

        Kn, Mn, Bn, Sn = self.triangularElement(node1, node2, node3)
        pp = array([phis[elms[e, 0]], phis[elms[e, 1]], phis[elms[e, 2]]])
        alpha = dot(Bn, pp) / (2. * Sn)
        phi = dot(r_[1., abpoint], alpha)
        phix = alpha[1]
        phiy = alpha[2]

        return phi, phix, phiy
    
    ## for judging inner point
    def isInner(self, po, p1, p2, p3):
        v1 = p1 - po
        v2 = p2 - po
        v3 = p3 - po

        c1 = cross(v1, v2)
        c2 = cross(v2, v3)
        c3 = cross(v3, v1)

        c1 = c1 / linalg.norm(c1)
        c2 = c2 / linalg.norm(c2)
        c3 = c3 / linalg.norm(c3)

        ret = 0

        if (c1 == c2).all():
            if (c1 == c3).all():
                ret = 1

        return ret


## for calculating 3-dimensional sound field with fem
class Fem3D:
    """A simple 3-dimensional fem class for calculating sound field"""

    def triangularElement(self, n1, n2, n3, order):
        if order==2:
            return self.triangularElement2nd( n1,n2,n3 )
                
        node = [n1, n2, n3]

        a = n2 - n1
        b = n3 - n2
        c = n1 - n3

        t1 = math.acos(dot(a, b) / linalg.norm(a) / linalg.norm(b))
        t2 = math.acos(dot(b, c) / linalg.norm(b) / linalg.norm(c))
        t3 = math.acos(dot(c, a) / linalg.norm(c) / linalg.norm(a))

        T1 = tan(pi / 2 - t1)
        T2 = tan(pi  /2 - t2)
        T3 = tan(pi / 2 - t3)

        s = (linalg.norm(a) + linalg.norm(b) + linalg.norm(c)) / 2
        S = sqrt(s * (s - linalg.norm(a)) * (s - linalg.norm(b))
                 * (s - linalg.norm(c)))
        #S = 0.5 * linalg.norm(a) * linalg.norm(b) * sin(t1)

        K = array([[T2 + T3, -T3, -T2],
                   [-T3, T1 + T3, -T1],
                   [-T2, -T1, T1 + T2]]) / 2

        M = array([[2., 1., 1.],
                   [1., 2., 1.],
                   [1., 1., 2.]]) * S / 12.

        return K,M,S
    
    def triangularElement2nd(self, n1, n2, n3):
        node = [n1, n2, n3]

        a = n2 - n1
        b = n3 - n2
        c = n1 - n3

        t1 = math.acos(dot(a, b) / linalg.norm(a) / linalg.norm(b))
        t2 = math.acos(dot(b, c) / linalg.norm(b) / linalg.norm(c))
        t3 = math.acos(dot(c, a) / linalg.norm(c) / linalg.norm(a))

        T1 = tan(pi / 2 - t1)
        T2 = tan(pi  /2 - t2)
        T3 = tan(pi / 2 - t3)

        s = (linalg.norm(a) + linalg.norm(b) + linalg.norm(c)) / 2
        S = sqrt(s * (s - linalg.norm(a)) * (s - linalg.norm(b))
                 * (s - linalg.norm(c)))
        #S = 0.5 * linalg.norm(a) * linalg.norm(b) * sin(t1)

#        #1,2,3,4,5,6
#        K = array([[3*(T2+T3),        T3,        T2,            0,        -4*T2,        -4*T3],
#                    [      T3, 3*(T1+T3),        T1,        -4*T1,            0,        -4*T3],
#                    [      T2,        T1, 3*(T1+T2),        -4*T1,        -4*T2,            0],
#                    [       0,     -4*T1,     -4*T1, 8*(T1+T2+T3),        -8*T3,        -8*T2],
#                    [   -4*T2,         0,     -4*T2,        -8*T3, 8*(T1+T2+T3),        -8*T1],
#                    [   -4*T3,     -4*T3,         0,        -8*T2,        -8*T1, 8*(T1+T2+T3)]])/6
#
        #1,2,3,6,4,5 (Toggle matrix for gmsh format)
        K = array([[3 * (T2 + T3), T3, T2, -4 * T3, 0, -4 * T2],
                   [T3, 3 * (T1 + T3), T1, -4 * T3, -4 * T1, 0],
                   [T2, T1, 3 * (T1 + T2), 0, -4 * T1, -4 * T2],
                   [-4 * T3, -4 * T3, 0, 8 * (T1 + T2 + T3), -8 * T2, -8 * T1],
                   [0, -4 * T1, -4 * T1, -8 * T2, 8 * (T1 + T2 + T3), -8 * T3],
                   [-4 * T2, 0, -4 * T2, -8 * T1, -8 * T3, 8 * (T1 + T2 + T3)]]
                  ) / 6

#        #1,2,3,4,5,6
#        M = array([[6.,-1.,-1.,-4., 0., 0.],
#                    [-1., 6.,-1., 0.,-4., 0.],
#                    [-1.,-1., 6., 0., 0.,-4.],
#                    [-4., 0., 0.,32.,16.,16.],
#                    [0.,-4., 0.,16.,32.,16.],
#                    [0., 0.,-4.,16.,16.,32.]])*S/180.
#
        #1,2,3,6,4,5 (Toggle matrix for gmsh format)
        M = array([[6., -1., -1., 0., -4., 0.],
                   [-1., 6., -1., 0., 0., -4.],
                   [-1., -1., 6., -4., 0., 0.],
                   [0., 0., -4., 32., 16., 16.],
                   [-4., 0., 0., 16., 32., 16.],
                   [0., -4., 0., 16., 16., 32.]]) * S / 180.

        return K,M,S

    def tetrahedralElement(self, n1, n2, n3, n4, order):
        if order == 2:
            return self.tetrahedralElement2nd(n1, n2, n3, n4)

        node = [n1, n2, n3, n4]
        a = [0., 0., 0., 0.]
        b = [0., 0., 0., 0.]
        c = [0., 0., 0., 0.]
        d = [0., 0., 0., 0.]

        for kk in range(4):
            k = mod(kk, 4)
            l = mod(kk + 1, 4)
            m = mod(kk + 2, 4)
            n = mod(kk + 3, 4)

            # Intermediate variables that consist a, b, c and d
            xl = node[l][0]
            xm = node[m][0]
            xn = node[n][0]

            yl = node[l][1]
            ym = node[m][1]
            yn = node[n][1]

            zl = node[l][2]
            zm = node[m][2]
            zn = node[n][2]

            ek = 1
            if (k == 1) | (k == 3):
                ek = -1

            ak = ek * (xl * (ym * zn - yn * zm) + xm * (yn * zl - yl * zn)
                       + xn * (yl * zm - ym * zl))
            bk = ek * (yl * (zn - zm) + ym * (zl - zn) + yn *(zm - zl))
            ck = ek * (zl * (xn - xm) + zm * (xl - xn) + zn *(xm - xl))
            dk = ek * (xl * (yn - ym) + xm * (yl - yn) + xn *(ym - yl))

            a[k] = ak
            b[k] = bk
            c[k] = ck
            d[k] = dk

        v = array([[1, node[0][0], node[0][1], node[0][2]],
                   [1, node[1][0], node[1][1], node[1][2]],
                   [1, node[2][0], node[2][1], node[2][2]],
                   [1, node[3][0], node[3][1], node[3][2]]])

        V = linalg.det(v) / 6.

        T12 = -1 / (6 * V) * (b[2] * b[3] + c[2] * c[3] + d[2] * d[3])
        T13 = -1 / (6 * V) * (b[1] * b[3] + c[1] * c[3] + d[1] * d[3])
        T14 = -1 / (6 * V) * (b[2] * b[1] + c[2] * c[1] + d[2] * d[1])

        T23 = -1 / (6 * V) * (b[0] * b[3] + c[0] * c[3] + d[0] * d[3])
        T24 = -1 / (6 * V) * (b[0] * b[2] + c[0] * c[2] + d[0] * d[2])

        T34 = -1 / (6 * V) * (b[0] * b[1] + c[0] * c[1] + d[0] * d[1])

        K = array([[T34 + T24 + T23, -T34, -T24, -T23],
                   [-T34, T34 + T14 + T13, -T14, -T13],
                   [-T24,-T14, T24 + T14 + T12, -T12],
                   [-T23, -T13, -T12, T23 + T13 + T12]]) / 6.

        M = array([[2., 1., 1., 1.],
                   [1., 2., 1., 1.],
                   [1., 1., 2., 1.],
                   [1., 1., 1., 2.]]) * V / 20.

        return K, M, V

    def tetrahedralElement2nd(self, n1, n2, n3, n4):
        node = [n1, n2, n3, n4]
        a = [0., 0., 0., 0.]
        b = [0., 0., 0., 0.]
        c = [0., 0., 0., 0.]
        d = [0., 0., 0., 0.]

        for kk in range(4):
            k = mod(kk, 4)
            l = mod(kk + 1, 4)
            m = mod(kk + 2, 4)
            n = mod(kk + 3, 4)

            # Intermediate variables that consist a, b, c and d
            xl = node[l][0]
            xm = node[m][0]
            xn = node[n][0]

            yl = node[l][1]
            ym = node[m][1]
            yn = node[n][1]

            zl = node[l][2]
            zm = node[m][2]
            zn = node[n][2]

            ek = 1
            if (k == 1) | (k == 3):
                ek = -1

            ak = ek * (xl * (ym * zn - yn * zm) + xm * (yn * zl - yl * zn)
                       + xn * (yl * zm - ym * zl))
            bk = ek * (yl * (zn - zm) + ym * (zl - zn) + yn * (zm - zl))
            ck = ek * (zl * (xn - xm) + zm * (xl - xn) + zn * (xm - xl))
            dk = ek * (xl * (yn - ym) + xm * (yl - yn) + xn * (ym - yl))

            a[k] = ak
            b[k] = bk
            c[k] = ck
            d[k] = dk

        v = array([[1, node[0][0], node[0][1], node[0][2]],
                   [1, node[1][0], node[1][1], node[1][2]],
                   [1, node[2][0], node[2][1], node[2][2]],
                   [1, node[3][0], node[3][1], node[3][2]]])

        V = linalg.det(v) / 6.

        Cx = 1. / (6.0 * V) * array([[4 * b[0], 0., 0., 0., -b[0]],
                                     [0., 4 * b[1], 0., 0., -b[1]],
                                     [0., 0., 4 * b[2], 0., -b[2]],
                                     [0., 0., 0., 4 * b[3], -b[3]],
                                     [4 * b[1], 4 * b[0], 0., 0., 0.],
                                     [0., 4 * b[2], 4 * b[1], 0., 0.],
                                     [4 * b[2], 0., 4 * b[0], 0., 0.],
                                     [4 * b[3], 0., 0., 4 * b[0], 0.],
                                     [0., 4 * b[3], 0., 4 * b[1], 0.],
                                     [0., 0., 4 * b[3], 4 * b[2], 0.]]).T

        Cy = 1. / (6.0 * V) * array([[4 * c[0], 0., 0., 0., -c[0]],
                                     [0., 4 * c[1], 0., 0., -c[1]],
                                     [0., 0., 4 * c[2], 0., -c[2]],
                                     [0., 0., 0., 4 * c[3], -c[3]],
                                     [4 * c[1], 4 * c[0], 0., 0., 0.],
                                     [0., 4 * c[2], 4 * c[1], 0., 0.],
                                     [4 * c[2], 0., 4 * c[0], 0., 0.],
                                     [4 * c[3], 0., 0., 4 * c[0], 0.],
                                     [0., 4 * c[3], 0., 4 * c[1], 0.],
                                     [0., 0., 4 * c[3], 4 * c[2], 0.]]).T

        Cz = 1. / (6.0 * V) * array([[4 * d[0], 0., 0., 0., -d[0]],
                                     [0., 4 * d[1], 0., 0., -d[1]],
                                     [0., 0., 4 * d[2], 0., -d[2]],
                                     [0., 0., 0., 4 * d[3], -d[3]],
                                     [4 * d[1], 4 * d[0], 0., 0., 0.],
                                     [0., 4 * d[2], 4 * d[1], 0., 0.],
                                     [4 * d[2], 0., 4 * d[0], 0., 0.],
                                     [4 * d[3], 0., 0., 4 * d[0], 0.],
                                     [0., 4 * d[3], 0., 4 * d[1], 0.],
                                     [0., 0., 4 * d[3], 4 * d[2], 0.]]).T
        
        NxNx = zeros((10, 10), 'float64')
        NyNy = zeros((10, 10), 'float64')
        NzNz = zeros((10, 10), 'float64')
        
        for i in range(10):
            for j in range(10):
                NxNx[i, j] \
                =  V / 10. * (Cx[0, i] * Cx[0, j] + Cx[1, i] * Cx[1, j]
                              + Cx[2, i] * Cx[2, j] + Cx[3, i] * Cx[3, j]) \
                + V / 20. * (Cx[0, i] * Cx[1, j] + Cx[0, i] * Cx[2, j]
                             + Cx[0, i] * Cx[3, j] + Cx[1, i] * Cx[0, j]
                             + Cx[1, i] * Cx[2, j] + Cx[1, i] * Cx[3, j]
                             + Cx[2, i] * Cx[0, j] + Cx[2, i] * Cx[1, j]
                             + Cx[2, i] * Cx[3, j] + Cx[3, i] * Cx[0, j]
                             + Cx[3, i] * Cx[1, j] + Cx[3, i] * Cx[2, j]) \
                + V / 4. * (Cx[0, i] * Cx[4, j] + Cx[1, i] * Cx[4, j]
                            + Cx[2, i] * Cx[4, j] + Cx[3, i] * Cx[4, j]
                            + Cx[4, i] * Cx[0, j] + Cx[4, i] * Cx[1, j]
                            + Cx[4, i] * Cx[2, j] + Cx[4, i] * Cx[3, j]) \
                + V * Cx[4, i] * Cx[4, j]

                NyNy[i, j] \
                =  V / 10. * (Cy[0, i] * Cy[0, j] + Cy[1, i] * Cy[1, j]
                              + Cy[2, i] * Cy[2, j] + Cy[3, i] * Cy[3, j]) \
                + V / 20. * (Cy[0, i] * Cy[1, j] + Cy[0, i] * Cy[2, j]
                             + Cy[0, i] * Cy[3, j] + Cy[1, i] * Cy[0, j]
                             + Cy[1, i] * Cy[2, j] + Cy[1, i] * Cy[3, j]
                             + Cy[2, i] * Cy[0, j] + Cy[2, i] * Cy[1, j]
                             + Cy[2, i] * Cy[3, j] + Cy[3, i] * Cy[0, j]
                             + Cy[3, i] * Cy[1, j] + Cy[3, i] * Cy[2, j]) \
                + V / 4.  * (Cy[0, i] * Cy[4, j] + Cy[1, i] * Cy[4, j]
                             + Cy[2, i] * Cy[4, j] + Cy[3, i] * Cy[4, j]
                             + Cy[4, i] * Cy[0, j] + Cy[4, i] * Cy[1, j]
                             + Cy[4, i] * Cy[2, j] + Cy[4, i] * Cy[3, j]) \
                + V * Cy[4, i] * Cy[4, j]
                NzNz[i, j] \
                =  V / 10. * (Cz[0, i] * Cz[0, j] + Cz[1, i] * Cz[1, j]
                              + Cz[2, i] * Cz[2, j] + Cz[3, i] * Cz[3, j]) \
                + V / 20. * (Cz[0, i] * Cz[1, j] + Cz[0, i] * Cz[2, j]
                             + Cz[0, i] * Cz[3, j] + Cz[1, i] * Cz[0, j]
                             + Cz[1, i] * Cz[2, j] + Cz[1, i] * Cz[3, j]
                             + Cz[2, i] * Cz[0, j] + Cz[2, i] * Cz[1, j]
                             + Cz[2, i] * Cz[3, j] + Cz[3, i] * Cz[0, j]
                             + Cz[3, i] * Cz[1, j] + Cz[3, i] * Cz[2, j]) \
                + V / 4. * (Cz[0, i] * Cz[4, j] + Cz[1, i] * Cz[4, j]
                            + Cz[2, i] * Cz[4, j] + Cz[3, i] * Cz[4, j]
                            + Cz[4, i] * Cz[0, j] + Cz[4, i] * Cz[1, j]
                            + Cz[4, i] * Cz[2, j] + Cz[4, i] * Cz[3, j]) \
                + V * Cz[4, i] * Cz[4, j]

        #1, 2, 3, 4, 5, 6, 7, 8, 9, 10
        K = NxNx + NyNy + NzNz
        #K = -K
        
        #1, 2, 3, 4, 5, 6, 7, 8, 9, 10
        M = array([[6., 1., 1., 1., -4., -6., -4., -4., -6., -6.],
                   [1., 6., 1., 1., -4., -4., -6., -6., -4., -6.],
                   [1., 1., 6., 1., -6., -4., -4., -6., -6., -4.],
                   [1., 1., 1., 6., -6., -6., -6., -4., -4., -4.],
                   [-4., -4., -6., -6., 32., 16., 16., 16., 16., 8.],
                   [-6., -4., -4., -6., 16., 32., 16., 8., 16., 16.],
                   [-4., -6., -4., -6., 16., 16., 32., 16., 8., 16.],
                   [-4., -6., -6., -4., 16., 8., 16., 32., 16., 16.],
                   [-6., -4., -6., -4., 16., 16., 8., 16., 32., 16.],
                   [-6., -6., -4., -4., 8., 16., 16., 16., 16., 32.], ]) \
                   * V / 420.

        M[8, :], M[9, :] = M[9, :].copy(), M[8, :].copy()
        M[:, 8], M[:, 9] = M[:, 9].copy(), M[:, 8].copy()
        
        K[8, :], K[9, :] = K[9, :].copy(), K[8, :].copy()
        K[:, 8], K[:, 9] = K[:, 9].copy(), K[:, 8].copy()

        return K, M, V

    def getAtArbitraryPoint(self, abpoint, phis, elms, nodes):
        elmnum = elms.shape[0]

        e = -1

        for N in range(elmnum):
            # Load the nodes that consist Nth element
            node1 = nodes[elms[N, 0], 0:2]
            node2 = nodes[elms[N, 1], 0:2]
            node3 = nodes[elms[N, 2], 0:2]

            if self.isInner(abpoint, node1, node2, node3):
                e = N
                break

        if e<0:
            return

        Kn, Mn, Bn, Sn = self.triElement(node1, node2, node3)
        pp = array([phis[elms[e, 0]], phis[elms[e, 1]], phis[elms[e, 2]]])
        alpha = dot(Bn, pp) / (2 * Sn)
        phi = dot(r_[1, abpoint], alpha)
        phix = alpha[1]
        phiy = alpha[2]

        return phi, phix, phiy

    ## for judging inner point
    def isInner(self, po, p1, p2, p3):
        v1 = p1 - po
        v2 = p2 - po
        v3 = p3 - po

        c1 = cross(v1, v2)
        c2 = cross(v2, v3)
        c3 = cross(v3, v1)

        c1 = c1 / linalg.norm(c1)
        c2 = c2 / linalg.norm(c2)
        c3 = c3 / linalg.norm(c3)

        ret = 0

        if (c1 == c2).all():
            if (c1 == c3).all():
                ret = 1

        return ret
