#  ooo  oooo  oooo o   o     o    ooo   ooo  o   o  oooo ooooo o  ooo   oooo
# 0   0 0   0 0    00  0    0 0  0   0 0   0 0   0 0       0   0 0   0 0
# 0   0 oooo  oooo 0 0 0   0   0 0     0   0 0   0  oooo   0   0 0      oooo
# 0   0 o     0    0 o 0   ooooo 0   o 0   0 0   0      0  0   0 0   o      0
#  ooo  o     oooo o  0o  0     0 ooo   ooo   ooo  0oooo   0   0  ooo  0oooo
#
#  (  (  (  (  (  (  (  (  (  (  (  (  O  )  )  )  )  )  )  )  )  )  )  )  )
#
#    OpenAcoustics: The Open Source Collection for Computational Acoustics
#
#               Copyright (C) 2008-2009 The OpenAcoustics Team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

## @package openacoustics.gmsh
# module for handling gmsh. 
# this call gmsh-excutable internally, and gmsh should be set to PATH.

__author__ = "hisaha <hisaha@souldsp.jp>"
__date__ = "2008/02/07"
__credits__ = "Hisharu Suzuki"
__version__ = "$Revision: 0 $"

from time import time

from scipy import *
from pylab import *
import math
import os
import sys

m_version = sys.version_info[0]
if  m_version == 3:
    import _pickle as pickle
elif m_version == 2:
    import cPickle as pickle
else:
    print("python major version error!!")
    exit()


## for parsing geo file.
class GeoParser:
    """A simple gmsh wrapper class for parsing geo file."""
    dim = 2
    
    ##visualize geometry by using gmsh
    #  @param self The object pointer.
    #  @param geo path to geometory file
    def viewGeo(self, geo):
        """visualize geometry by using gmsh"""
        os.system(str('gmsh ' + geo))

    ##load geometry file to instance
    #  @param self The object pointer.
    #  @param geo path to geometory file
    #  @param clscale scale of element. (In the case of less than 1, good calculation accuracy but require PC resource)
    #  @param order order of element
    def loadGeo(self, geo, clscale = 1, order = 1):
        """load geometry file to instance"""
        os.system(str('gmsh ' + geo + ' -' + str(self.dim) + ' -clscale '
                      + str(clscale) + ' -order ' + str(order)  +  ' -format msh2 -o tmp.msh'))
        f = open('tmp.msh', 'r')
        self.msh = f.readlines()
        f.close()

    ##load mesh file to instance
    #  @param self The object pointer.
    #  @param msh path to mesh file
    def loadMsh(self, msh):
        """load mesh file to instance"""
        f = open(msh, 'r')
        self.msh = f.readlines()
        f.close()

    ## read nodes from loaded geo file
    #  @param self The object pointer
    #  @retval nodes loaded nodes    
    def getNodes(self):
        """read nodes from loaded geo file."""
        nodes = self.read('Nodes', 'EndNodes')
        #nodes = nodes[:, 1:4]
        #return cast['float64'](nodes)
        if self.dim == 2:
            nodes = nodes[:, -3:-1]
        if self.dim == 3:
            nodes = nodes[:, -3:]

        return nodes

    ## read point elements with elmtype(int) and physname(str)
    #  @param self The object pointer
    #  @param physname physical name in loaded geo file
    #  @retval elm point elements of physical name
    def getPoints(self, physname):
        """read point elements with elmtype(int) and physname(str)"""
        elm = self.getElements(15, physname)
        return elm[:, -1]

    ## read triangular elements with elmtype(int) and physname(str)
    #  @param self The object pointer
    #  @param physname physical name in loaded geo file
    #  @param order order of element
    #  @retval triangular elements of physical name
    def getTriangles(self, physname, order = 1):
        """read triangular elements with elmtype(int) and physname(str)"""
        if order == 1:
            elm = self.getElements(2, physname)
            return elm[:, -3:]
        elif order == 2:
            elm = self.getElements(9, physname)
            return elm[:, -6:]
            #elm = elm[:, -6:]
            #return c_[elm[:, 0:3], elm[:, 4], elm[:, 5], elm[:, 3]]

    ## read tetrahedron elements with elmtype(int) and physname(str)
    #  @param self The object pointer
    #  @param physname physical name in loaded geo file
    #  @param order order of element
    #  @retval elm tetrahedron elements of physical name
    def getTetrahedra(self, physname, order = 1):
        """read tetrahedron elements with elmtype(int) and physname(str)"""
        if order == 1:
            elm = self.getElements(4, physname)
            return elm[:, -4:]
        elif order == 2:
            elm = self.getElements(11, physname)
            return elm[:, -10:]
    
    ## read line elements with elmtype(int) and physname(str)
    #  @param self The object pointer
    #  @param physname physical name in loaded geo file
    #  @retval elm line elements of physical name
    def getLines(self, physname, order=1):
        """read line elements with elmtype(int) and physname(str)"""
        if order == 1:        
            elm = self.getElements(1, physname)
            return elm[:, -2:]
        elif order == 2:
            elm = self.getElements(8, physname)
            return elm[:, -3:]
        

    ## read arbitrary elements with elmtype(int) and physname(str)
    #  @param self The object pointer
    #  @param elmtype element type (documented at gmsh web)
    #  @param physname physical name in loaded geo file
    #  @retval elm rawdata of elements of physical name
    def getElements(self, elmtype, physname):
        """read arbitrary elements with elmtype(int) and physname(str)"""
        lines = self.findlines("Elements", "EndElements", elmtype, physname)
        elm = self.readfromlines(lines)
        elm = elm - 1
        return cast["int32"](elm)

    ## call internally. get id of physical name. modified by Auis Abe, and Tsukui. Thanks
    #  @param self The object pointer
    #  @param name physical name
    #  @retval id id of physical name
    def getnameid( self, name ):
        idxs = self.findbetween( "PhysicalNames","EndPhysicalNames" )
        if not idxs: return -1
        names = self.msh[ idxs[0]:idxs[1] ]
        id=-1
        if len(name)>0:
            version = self.getMeshFormat()[0]
            if version <= "2.0":
                for tmp in names:
                    tmp = tmp.split()
                    if tmp[1] == str( "\""+str(name)+"\"" ):
                        id=int( tmp[0] )
            else:
                for tmp in names:
                    tmp = tmp.split()
                    if tmp[2] == str( "\""+str(name)+"\"" ):
                        id=int( tmp[1] )                
            #else:
            #    raise ValueError, "MeshFileFormat Error."
        return id

    ## get file format. written by Auis Abe, and Tsukui. Thanks
    #  @param self The object pointer
    #  @retval versionNumber, fileType, dataSize
    def getMeshFormat(self):
        format = self.msh[1]
        return format.split() # [versionNumber, fileType, dataSize]


    ## call internally.
    def readfromlines(self, lines):
        tmp = self.msh[lines[0]].split()
        elms = zeros((len(lines), len(tmp)), "float64")

        for i in range(len(lines)):
            tmp = self.msh[lines[i]].split()
            for j in range(0, len(tmp)):
                elms[i, j] = float(tmp[j])

        return elms

    ## call internally. 
    def readbetween(self, line1, line2):
        lines = range(line1, line2)
        return self.readfromlines(lines)

    ## call internally.
    def read(self, str1, str2):
        idxs = self.findbetween(str1, str2)
        return self.readbetween(idxs[0], idxs[1])

    ## call internally.
    def findbetween(self, str1, str2):
        s = 0
        idxs = []
        for i in range(len(self.msh)):
            if (self.msh[i].find(str1) != -1) & (s == 0):
                idxs.append(i + 2)
                s = 1
            elif (self.msh[i].find(str2) != -1) & (s == 1):
                idxs.append(i)

        return idxs

    ## call internally.
    def findlines(self, str1, str2, tp, nm):
        idxs = self.findbetween(str1, str2)
        if str(nm)==nm:
            physid = self.getnameid(nm)
        else:
            physid = nm
        lines = []
        if physid > 0:
            for i in range(idxs[0], idxs[1]):
                tmp = self.msh[i].split()
                #print self.msh[i][3]
                if (int(tmp[1]) == tp) & (int(tmp[3]) == physid):
                    lines.append(i)
        else:
            for i in range(idxs[0], idxs[1]):
                tmp = self.msh[i].split()
                if (int(tmp[1]) == tp):
                    lines.append(i)

        return lines

## for creating postprocessing file
class PostProcessor:
    """A simple gmsh wrapper class for creating postprocessor file.
    This class is highly experimental."""
    tmpfile = ""
    
    ##declare begining of post processing
    #  @param self The object pointer.
    def beginPostprocessing(self):
        """declare begining of post processing"""
        self.tmpfile = open('tmpr', 'w')
    
    ##declare ending of post processing
    #  @param self The object pointer.
    def endPostprocessing(self):
        """declare ending of post processing"""
        self.tmpfile.close()

    ##declare run visualising
    #  @param self The object pointer.
    def run(self):
        """run visualising"""
        os.system('gmsh tmpr')

    def merge(self, fname):
        self.tmpfile.write("Merge \"" + str(fname) + "\";\n")

    def addStyle(self):
        self.tmpfile.write("General.BackgroundGradient = 0;\n")
        self.tmpfile.write("General.Color.Background = {255, 255, 255};\n")
        self.tmpfile.write("General.Color.Text = {0, 0, 0};\n")
        self.tmpfile.write("Print.EpsBackground = 0;\n")

    def beginView(self, name):
        self.tmpfile.write(str("View \"" + name + "\"{\n"))

    def closeView(self):
        self.tmpfile.write("};\n")

    def title(self, vid, t):
        self.tmpfile.write(str("View[0].Name = \"" + t + "\";\n"))

    def xLabel(self, vid, xlabel):
        self.tmpfile.write(str("View[" + str(vid) + "].AxesLabelX = \""
                               + xlabel + "\";\n"))
        self.tmpfile.write("View[" + str(vid) + "].Axes = 1;\n")
        self.tmpfile.write("View[" + str(vid) + "].Color.Axes = {0, 0, 0};\n")

    def yLabel(self, vid, ylabel):
        self.tmpfile.write(str("View[0].AxesLabelY = \"" + ylabel + "\";\n"))
        self.tmpfile.write("View[" + str(vid) + "].Axes = 1;\n")
        self.tmpfile.write("View[" + str(vid) + "].Color.Axes = {0, 0, 0};\n")

    def zLabel(self, vid, zlabel):
        self.tmpfile.write(str("View[0].AxesLabelZ = \"" + zlabel + "\";\n"))
        self.tmpfile.write("View[" + str(vid) + "].Axes = 1;\n")
        self.tmpfile.write("View[" + str(vid) + "].Color.Axes = {0, 0, 0};\n")

    def addScalarLine(self, f, x, y, z, s):
        self.addScalar(self.tmpfile, "SL", x, y, z, s)

    def addScalarTriangle(self, x, y, z, s):
        self.addScalar(self.tmpfile,"ST", x, y, z, s)

    def addScalarTriangle2nd(self, x, y, z, s):
        self.addScalar(self.tmpfile, "ST2", x, y, z, s)

    def addScalarTetrahedron(self, x, y, z, s):
        self.addScalar(self.tmpfile, "SS", x, y, z, s)

    def addScalarTetrahedron2nd(self, x, y, z, s):
        self.addScalar(self.tmpfile, "SS2", x, y, z, s)

    def addScalar(self, f, tp, x, y, z, s):
        f.write(str(tp+"("))
        tmp = ""
        for j in range(len(x)):
            tmp = tmp + str(x[j]) + ", " + str(y[j]) + ", " + str(z[j])
            if j != (len(x) - 1):
                tmp = tmp + ", "

        f.write(tmp)
        self.tmpfile.write("){")
        tmp = ""

        for j in range(len(s)):
            tmp = tmp + str(s[j])
            if j != (len(s) - 1):
                tmp = tmp + ", "

        f.write(tmp)
        f.write("};\n")

    def addCutPlane(self, view, p1, p2, p3):
        v1 = p2 - p1
        v2 = p3 - p1

        ABC = cross(v1, v2)
        D = - dot(ABC, p1)

        self.addCutPlaneByABCDForm(view, ABC[0], ABC[1], ABC[2], D)

    def addCutPlaneByABCDForm(self, view, A, B, C, D):
        #self.tmpfile.write("View[" + str(view) + "].ColormapNumber = 1;\n")
        #self.tmpfile.write("View[1].ColormapNumber = 0;\n")
        self.tmpfile.write("Plugin(CutPlane).iView = " + str(view) + ";\n")
        self.tmpfile.write("Plugin(CutPlane).A = " + str(A) + ";\n")
        self.tmpfile.write("Plugin(CutPlane).B = " + str(B) + ";\n")
        self.tmpfile.write("Plugin(CutPlane).C = " + str(C) + ";\n")
        self.tmpfile.write("Plugin(CutPlane).D = " + str(D) +";\n")
        self.tmpfile.write("Plugin(CutPlane).Run;\n")

## for handling 2-dimensinal geometry and nodes
class GMsh2D(GeoParser, PostProcessor):
    """A simple gmsh wrapper class for parsing 2-dimensional geo file.
    This class only treat 2-dimensinoal nodes and points, lines and triangular elements.
    """
    
    dim = 2
    
    ##visualize contour of triagnel meshes by using gmsh.
    #  The cordinates of vertices are x, y, z arrays and the vertices are indexed by idxs array.
    #  @param self The object pointer.
    #  @param idxs index array of x,y,z
    #  @param vertices of x
    #  @param vertices of y
    #  @param vertices of z
    #  @param title title of graph
    #  @param xlabel label of x axis
    #  @param ylabel label of y axis
    #  @param order order of element
    def contour(self, idxs, x, y, z, title, xlabel, ylabel, order = 1,
                file = ""):
        """visualize contour of triagnel meshes by using gmsh.
        The cordinates of vertices are x, y, z arrays and the vertices are indexed by idxs array.
        """
        self.beginPostprocessing()
        self.beginView("contour")

        xx = x.max() - x.min()
        yy = y.max() - y.min()
        zz = z.max() - z.min()

        L = max(xx, yy)

        if order == 1:
            for i in range(idxs.shape[0]):
                X = []
                Y = []
                Z = []
                S = []
                for j in range(idxs.shape[1]):
                    X.append(x[idxs[i, j]])
                    Y.append(y[idxs[i, j]])
                    Z.append(z[idxs[i, j]] / zz * L / 1.6)
                    S.append(z[idxs[i, j]])
                self.addScalarTriangle(X, Y, Z, S)
        elif order == 2:
            for i in range(idxs.shape[0]):
                X = []
                Y = []
                Z = []
                S = []
                for j in range(idxs.shape[1]):
                    X.append(x[idxs[i, j]])
                    Y.append(y[idxs[i, j]])
                    Z.append(z[idxs[i, j]] / zz * L / 1.6)
                    S.append(z[idxs[i, j]])
                self.addScalarTriangle2nd(X, Y, Z, S)

        self.closeView()
        self.addStyle()

        self.title(0, title)
        self.xLabel(0, xlabel)
        self.yLabel(0, ylabel)
        self.tmpfile.write("View[0].OffsetY = "
                           + str(0.1 * (y.max() - y.min())) + ";\n")
        if not (file == ""):
            self.tmpfile.write("Print \"" + file + "\";\n")
        self.endPostprocessing()
        self.run()

## for handling 3-dimensinal geometry and nodes
class GMsh3D(GeoParser, PostProcessor):
    """A simple gmsh wrapper class for parsing 3-dimensional geo file.
    This class only treat 3-dimensinoal nodes and points, lines and tetrahedoron elements.
    """
    dim = 3

    def addTetrahedron(self, idxs, x, y, z, scalars):
        for i in range(idxs.shape[0]):
            X = []
            Y = []
            Z = []
            Sc = []
            for j in range(idxs.shape[1]):
                X.append(x[idxs[i, j]]) # = tmp + str(x[idxs[i, j]]) + ", "
                # + str(y[idxs[i, j]]) + ", " + str(z[idxs[i, j]])
                Y.append(y[idxs[i, j]])
                Z.append(z[idxs[i, j]])
                Sc.append(scalars[idxs[i, j]])
            self.addScalarTetrahedron(X, Y, Z, Sc)

    def addTetrahedron2nd(self, idxs, x, y, z, scalars):
        for i in range(idxs.shape[0]):
            X = []
            Y = []
            Z = []
            Sc = []
            for j in range(idxs.shape[1]):
                X.append(x[idxs[i, j]]) # = tmp + str(x[idxs[i, j]]) + ", "
                # + str(y[idxs[i, j]]) + ", " + str(z[idxs[i, j]])
                Y.append(y[idxs[i, j]])
                Z.append(z[idxs[i, j]])
                Sc.append(scalars[idxs[i, j]])
            self.addScalarTetrahedron2nd(X, Y, Z, Sc)

    def plotTriangularElements(self, idxs, x, y, z):
        self.beginPostprocessing()
        self.beginview("triangles")
        for i in range(idxs.shape[0]):
            X = [x[idxs[i, 0]], x[idxs[i, 1]]]
            Y = [y[idxs[i, 0]], y[idxs[i, 1]]]
            Z = [z[idxs[i, 0]], z[idxs[i, 1]]]
            S = [0., 0.]
            self.addScalarLine(self.tmpfile, X, Y, Z, S)

            X = [x[idxs[i, 2]], x[idxs[i, 1]]]
            Y = [y[idxs[i, 2]], y[idxs[i, 1]]]
            Z = [z[idxs[i, 2]], z[idxs[i, 1]]]
            S = [0., 0.]
            self.addScalarLine(self.tmpfile, X, Y, Z, S)

            X = [x[idxs[i, 0]], x[idxs[i, 2]]]
            Y = [y[idxs[i, 0]], y[idxs[i, 2]]]
            Z = [z[idxs[i, 0]], z[idxs[i, 2]]]
            S = [0., 0.]
            self.addScalarLine(self.tmpfile, X, Y, Z, S)

        self.closeview()
        self.addstyle()
        self.endPostprocessing()
        self.run()

    ## visualize contour3 of triagular meshes by using gmsh.
    #  The cordinates of vertices are x, y, z arrays and the vertices are indexed by idxs array.
    #  @param self The object pointer.
    #  @param idxs index array of x,y,z
    #  @param vertices of x
    #  @param vertices of y
    #  @param vertices of z
    #  @param scalar values at x,y,z
    #  @param title title of graph
    #  @param xlabel label of x axis
    #  @param ylabel label of y axis
    #  @param order order of element
    def contour3(self, idxs, x, y, z, scalars, title, xlabel, ylabel, order):
        """visualize contour3 of triagular meshes by using gmsh.
        The cordinates of vertices are x, y, z arrays and the vertices are indexed by idxs array.
        """
        self.beginPostprocessing()
        self.beginView("contour3")
        if order == 2:
            self.addTetrahedron2nd(idxs, x, y, z, scalars)
        elif order == 1:
            self.addTetrahedron(idxs, x, y, z, scalars)
            
        self.closeView()
        self.addStyle()

        self.title(0, title)
        self.xLabel(0, xlabel)
        self.yLabel(0, ylabel)

        self.endPostprocessing()
        self.run()
