#  ooo  oooo  oooo o   o     o    ooo   ooo  o   o  oooo ooooo o  ooo   oooo
# 0   0 0   0 0    00  0    0 0  0   0 0   0 0   0 0       0   0 0   0 0
# 0   0 oooo  oooo 0 0 0   0   0 0     0   0 0   0  oooo   0   0 0      oooo
# 0   0 o     0    0 o 0   ooooo 0   o 0   0 0   0      0  0   0 0   o      0
#  ooo  o     oooo o  0o  0     0 ooo   ooo   ooo  0oooo   0   0  ooo  0oooo
#
#  (  (  (  (  (  (  (  (  (  (  (  (  O  )  )  )  )  )  )  )  )  )  )  )  )
#
#    OpenAcoustics: The Open Source Collection for Computational Acoustics
#
#               Copyright (C) 2008-2012 The OpenAcoustics Team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.



# import modules 
from numpy import array, linalg, zeros, dot, pi, sqrt
from scipy import special
from scipy.integrate import quadrature


#=====================================================================
#                        Definitions of functions
#=====================================================================
# Hankel functions
def h01(x):
    return special.j0(x) + 1j * special.y0(x)

def h11(x):
    return special.j1(x) + 1j * special.y1(x)

# Green function of the 2-dimensional free field
def GreenFunction(rp, rs, waveNumber):
    rp_rs = rs - rp
    r = sqrt(rp_rs[0] ** 2. + rp_rs[1] ** 2.)

    return (1j / 4.) * h01(r * waveNumber)

# Integrands of element integral for h and g
def integrand_h(xi, element, rp, waveNumber):
    phi1 = (1. - xi) / 2.
    phi2 = (1. + xi) / 2.
    xr = phi1 * element['r1'][0] + phi2 * element['r2'][0]
    yr = phi1 * element['r1'][1] + phi2 * element['r2'][1]
    xp_xr = rp[0] - xr
    yp_yr = rp[1] - yr
    r = sqrt(xp_xr ** 2. + yp_yr ** 2.)

    r2_r1 = element['r2'] - element['r1']
    normalVector = array([ r2_r1[1], - r2_r1[0] ]) \
            / sqrt(r2_r1[0] ** 2. + r2_r1[1] ** 2.)
    dr_dn = - (xp_xr * normalVector[0] + yp_yr * normalVector[1]) / r

    return h11(r * waveNumber) * dr_dn

def integrand_g(xi, element, rp, waveNumber):
    phi1 = (1. - xi) / 2.
    phi2 = (1. + xi) / 2.
    xr = phi1 * element['r1'][0] + phi2 * element['r2'][0]
    yr = phi1 * element['r1'][1] + phi2 * element['r2'][1]
    xp_xr = rp[0] - xr
    yp_yr = rp[1] - yr
    r = sqrt(xp_xr ** 2. + yp_yr ** 2.)

    return h01(r * waveNumber)


#=====================================================================
#                              Main routine
#=====================================================================
#========================================================== Initialize 
#-- sound speed -- assuming temperature = 15 in Celsius degree.
soundSpeed = 340.

#-- frequency --
frequency = 500.
angularFrequency = 2. * pi * frequency
waveNumber = angularFrequency / soundSpeed

#-- sources & observing points --
rs = array( [0.2, 0.2] )

rp = array( [ [-0.2,  0.2],
              [-0.2, -0.2],
              [ 0.2, -0.2] ] )

#-- boundary elements --
nElement = 4
element = [{ 'r1': array([.5 , .5 ]), 'r2': array([-.5, .5 ]), 
                'admittance': 0. },
           { 'r1': array([-.5, .5 ]), 'r2': array([-.5, -.5]),
               'admittance': 0. },
           { 'r1': array([-.5, -.5]), 'r2': array([.5,  -.5]), 
               'admittance': 1. },
           { 'r1': array([.5 , -.5]), 'r2': array([.5,  .5 ]), 
               'admittance': 0. }
          ]

#======================================== Build coefficient matrix 'A'
A = zeros((nElement,nElement), 'complex64')

# for each boundary element
for m in range(nElement): 
    r2_r1 = element[m]['r2'] - element[m]['r1']
    elementLength = sqrt(r2_r1[0] ** 2. + r2_r1[1] ** 2.)
    a = -1j * waveNumber * elementLength / (4. * 2.)
    b = waveNumber * elementLength \
            * element[m]['admittance'] / (4. * 2.)

    # for each middle point of element
    for l in range(nElement): 
        rl = (element[l]['r1'] + element[l]['r2']) / 2.

        if l != m: # general component
            h = a * quadrature(integrand_h, -1., 1.,
                    args=(element[m], rl, waveNumber))[0]
            g = b * quadrature(integrand_g, -1., 1.,
                    args=(element[m], rl, waveNumber))[0]
            A[l,m] = h + g

        else: # diagonal component 
            v = special.itj0y0(waveNumber * elementLength / 2.)
            g = (element[l]['admittance'] / 2.) * complex(v[0], v[1])
            A[l,m] = 1./2. +  g

# show A
print( 'Coefficient matrix A:')
print(A)
input('\nPress ENTER to continue\n')

#======================================== Build right hand vector 'pd'
pd = zeros((nElement,1), 'complex64')

for l in range(nElement):
    rl = (element[l]['r1'] + element[l]['r2']) / 2.
    pd[l] = GreenFunction( rl, rs, waveNumber )

#show pd
print( 'Right hand side vector pd:')
print(pd)
input('\nPress ENTER to continue\n')

#=================================== Solve linear equations A * p = pd
p = linalg.solve( A, pd )

#show p
print('Solution of linear equation p (A*p=pd):')
print(p)
input('\nPress ENTER to continue\n')

#====================================== Calculate sound pressure at rp
prp = zeros((rp.shape[0],1), 'complex64' )
Ar = zeros(nElement, 'complex64')

# for each observing point
for k in range(rp.shape[0]): 
    # for each boundary element
    for m in range(nElement): 
        r2_r1 = element[m]['r2'] - element[m]['r1']
        elementLength = sqrt(r2_r1[0] ** 2. + r2_r1[1] ** 2.)
        a = -1j * waveNumber * elementLength / (4. * 2.)
        b = waveNumber * elementLength \
                * element[m]['admittance'] / (4. * 2.)

        h = a * quadrature(integrand_h, -1., 1.,
                args=(element[m], rp[k,:], waveNumber))[0]
        g = b * quadrature(integrand_g, -1., 1.,
                args=(element[m], rp[k,:], waveNumber))[0]
        Ar[m] = h + g

    prp[k] = GreenFunction(rp[k,:], rs, waveNumber) - dot(Ar, p)[0]

#show results
print('Sound pressure at rp1, rp2 and rp3:')
print(prp)
input('\nPress ENTER to exit\n')

# End






