#  ooo  oooo  oooo o   o     o    ooo   ooo  o   o  oooo ooooo o  ooo   oooo
# 0   0 0   0 0    00  0    0 0  0   0 0   0 0   0 0       0   0 0   0 0
# 0   0 oooo  oooo 0 0 0   0   0 0     0   0 0   0  oooo   0   0 0      oooo
# 0   0 o     0    0 o 0   ooooo 0   o 0   0 0   0      0  0   0 0   o      0
#  ooo  o     oooo o  0o  0     0 ooo   ooo   ooo  0oooo   0   0  ooo  0oooo
#
#  (  (  (  (  (  (  (  (  (  (  (  (  O  )  )  )  )  )  )  )  )  )  )  )  )
#
#    OpenAcoustics: The Open Source Collection for Computational Acoustics
#
#               Copyright (C) 2008-2012 The OpenAcoustics Team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from numpy import array, dot, float64, linspace, pi, sqrt, zeros, ones, random, mod, cross
from math import cos, sin, acos
from scipy.linalg import norm
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot
import matplotlib.cm
import time

#--------------------------------------------------------------------------------
#                       1. reading data fuction series
#--------------------------------------------------------------------------------
####################################
#### 1.1 read room data         ####
####################################
# @param 1 : file name of room data 
#
def readRoomData( roomDataFileName ):
    file = open(roomDataFileName, "r")

    ### read point data
    global numberOfPoints
    global point

    numberOfPoints = int( file.readline() )
    print("number of point: ", numberOfPoints)
    point = zeros( (numberOfPoints,3), "float64")

    for m in range(numberOfPoints):
        itemList = file.readline().split(',')
        for n in range(len(itemList)):
            if(n == 0):
                num = int(itemList[n])
            else:
                point[m][n-1] = float64(itemList[n])
    
    ### read wall data
    global numberOfWalls
    global absorpCoeffIndex
    global numberOfApexes
    global wall
  
    numberOfWalls = int(file.readline())
    numberOfApexes = zeros(numberOfWalls, "int")
    wall = numberOfWalls*[0]
    absorpCoeffIndex = zeros(numberOfWalls, "int")

    for m in range(numberOfWalls):
        itemList = file.readline().split(',')
        wall[m] =  (len(itemList) - 3)*[0] 
        for n in range(len(itemList)):
            if(n == 0):
                wallIndex = int(itemList[n])
            elif(n == 1):
                absorpCoeffIndex[m] = int(itemList[n])
            elif(n == 2):
                numberOfApexes[m] = int(itemList[n])
            else:
                wall[m][n-3] = int(itemList[n])
    file.close()

###############################################
#### 1.2 read absorption coefficient data  ####
###############################################
# @param 1 :file name of absorption coefficient
#
def readAbsorpCoeffData( absorpCoeffDataFileName ):
    file = open(absorpCoeffDataFileName, "r")

    global numberOfMaterials
    global absorpCoeff

    numberOfMaterials = int(file.readline())

    absorpCoeff = zeros((numberOfMaterials,6), "float64")

    for m in range(numberOfMaterials):
        itemList = file.readline().split(',')
        for n in range(6):
            absorpCoeff[m][n] = float64(itemList[n+2])

    file.close()

##############################################
#### 1.3 read calculation condtion data   ####
##############################################
# @param 1: file name of calculation condition
#
def readCalcCondData( calculationConditionFileName ):
    file  = open(calculationConditionFileName, "r")

    global soundSpeed
    global soundSourcePosition
    global numberOfRays
    global receivePosition
    global radiusReceivingPoint
    global limitNumOfReflectTimes
    global totalTimeOfCalc
    global echoTimeInterval
    global timeInterval

    soundSpeed = float(file.readline().split(',')[1])

    soundSourcePosition = zeros(3, "float64")
    itemList = file.readline().split(',')[1:4]
    for n in range(3):
        soundSourcePosition[n] = float64(itemList[n]);
    
    numberOfRays = int(file.readline().split(',')[1])
    
    receivePosition = zeros(3, "float64")
    itemList = file.readline().split(',')[1:4]
    for n in range(3):
        receivePosition[n] = float64(itemList[n]);

    radiusReceivingPoint = float64(file.readline().split(',')[1])

    limitNumOfReflectTimes = int(file.readline().split(',')[1])
                               
    totalTimeOfCalc = float64(file.readline().split(',')[1])

    echoTimeInterval = float64(file.readline().split(',')[1])

    timeInterval = float64(file.readline().split(',')[1])


#--------------------------------------------------------------------------------
#2. geometrical function series for ray tracing
#--------------------------------------------------------------------------------
##########################################
##### 2.1 vectorization from 2 points ####
##########################################
# @param p1: point 1 (x1, y1, z1)
# @param p2: point 2 (x2, y2, z2)
#
def pointToVector( p1, p2 ):
    v = array(p2 - p1)
    return v

######################################
#### 2.2 calculate plane equation ####
######################################
#
def calcPlaneEq():
    global planeEquation
    planeEquation = zeros((numberOfWalls,4), "float64")
    for m in range(numberOfWalls):
        p1 = point[wall[m][0]]
        p2 = point[wall[m][1]]
        p3 = point[wall[m][2]]

        v1 = pointToVector(p1, p2)
        v2 = pointToVector(p3, p2)

        normalVector = cross(v1, v2)
        normalVector /= norm(normalVector)
        planeEquation[m][0:3] = normalVector
        d = (-1.0) * (normalVector[0]*p1[0] \
                    + normalVector[1]*p1[1] \
                    + normalVector[2]*p1[2])
        planeEquation[m][3] = d
        print(planeEquation[m])

###############################################################
#### 2.3 calculate cross point from one plane and one line ####
###############################################################
# @param 1 v: vector of sound ray (vx, vy, vz)
# @param 2 p: point on sound ray (x1, y1, z1)
# @param 3 pe: plane equation coefficient (a, b, c, d)
#
def calcCrossPoint(v, p, pe):
    crossPoint = zeros(3, "float64")
    a, b, c, d = pe
    t = -(a*p[0] + b*p[1] + c*p[2] + d) / (a*v[0] + b*v[1] + c*v[2])
    for n in range(3):
        crossPoint[n] = p[n] + t*v[n]
    return crossPoint

################################################
#### 2.4 calculate the distance of 2 points ####
################################################
# @param p1: point1 (x1, y1, z1)
# @param p2: point2 (x2, y2, z2)
# 
def calcPointToPointDistance(p1, p2):
    d = norm(p2 - p1)
    return d

###########################################################
#### 2.5 calculate the distance from a point to a line ####
###########################################################
# @param v: the vector of a line (vx, vy, vz)
# @param p1: a point on a line (x1, y1, z1)
# @param p2: a point away from a line (x2, y2, z2)
#
def calcPointToLineDistance(v, p1, p2): 
    t =  dot(v,(p2-p1)) / dot(v,v)
    h = t * v + p1               ## cross point: h(x, y, z)
    d = calcPointToPointDistance(h, p2)
    return d


#################################
#### 2.6 create sound source  ###
#################################
#
def createOmniSoundSource():
    global rayVector
    global rayPosition
    global actualNumberOfRays
    rayVector = zeros((numberOfRays,3), "float64")
    rayPosition = zeros((numberOfRays,3), "float64")
    d_omega = 4.0*pi/numberOfRays
    d_phi = sqrt(2.0*d_omega/sqrt(3.0))
    d_theta = sqrt(3.0)*d_phi/2.0
    theta = -pi/2.0
    actualNumberOfRays = 0
    counter = 0;
    while(theta <= pi/2.0):
        numberOfRaysOn_H_plane = int(round(cos(theta)*2.0*pi/d_phi))
        if(actualNumberOfRays + numberOfRaysOn_H_plane <= numberOfRays):
            for n in range(actualNumberOfRays, 
                           actualNumberOfRays + numberOfRaysOn_H_plane):
                if(mod(counter,2) == 0):
                    m = n
                else:
                    m = n + 1/2
                rayVector[n][0] = cos(2*pi*m/numberOfRaysOn_H_plane)*cos(theta)
                rayVector[n][1] = sin(2*pi*m/numberOfRaysOn_H_plane)*cos(theta)
                rayVector[n][2] = sin(theta)
                rayPosition[n] = soundSourcePosition
            actualNumberOfRays += numberOfRaysOn_H_plane
        if(actualNumberOfRays == numberOfRays):
            break
        theta += d_theta
        counter += 1
    print('Actual number of Sound Rays: ', actualNumberOfRays)
    return actualNumberOfRays

#-------------------------------------------------------------------------------------
#                   3. collision detection function series
#-------------------------------------------------------------------------------------
######################################################
### 3.1 Is a sound ray toward or backward a wall? ####
######################################################
# @param v1: vector of a sound ray
# @param v2: normal vector of a wall
#   
def isTowardWall(v1, v2):
    if(dot(v1, v2) < 0.0):
        return True # toward
    else:
        return False # backward

######################################
### 3.2 Is a point inside a wall? ####
######################################
# @param p1: a point on a plane
# @param wallIndex: Index of a wall
#
def isInsideWall(p1, wallIndex):
    totalDegreeOfAngles = 0.0
    for m in range(numberOfApexes[wallIndex] - 1):
        v1 = pointToVector(point[wall[wallIndex][m]], p1)
        v2 = pointToVector(point[wall[wallIndex][m+1]], p1)
        angle = acos( dot(v1,v2) / (norm(v1)*norm(v2)) )
        isClockwise = dot( cross(v1,v2), planeEquation[wallIndex][0:3] )
        if(isClockwise > 0.0):
               angle *= -1
        totalDegreeOfAngles += angle
    v1 = pointToVector(point[wall[wallIndex][numberOfApexes[wallIndex] - 1]], p1)
    v2 = pointToVector(point[wall[wallIndex][0]], p1)
    angle = acos( dot(v1,v2) / (norm(v1)*norm(v2)) )
    isClockwise = dot( cross(v1,v2), planeEquation[wallIndex][0:3] )
    if( isClockwise > 0.0):
        angle *= -1
    totalDegreeOfAngles += angle
    if(totalDegreeOfAngles > 2.0*pi*0.99 and \
       totalDegreeOfAngles < 2.0*pi*1.01):    
                            # 0.99 and 1.01 is magic number
        return True
    else:
        return False

#########################################
### 3.3 calculate reflection vector #####
#########################################
# @param vi: incident vector of line
# @param vn: normal vector of plane
#
def calcReflectionVector(vi, vn):
    reflectionVector = vi - 2 * dot(vn,vi) * vn
    reflectionVector  /= norm(reflectionVector)
    return reflectionVector

#---------------------------------------------------------------------------
#                                4. main routine
#---------------------------------------------------------------------------
print("#1 Read Room Data")
readRoomData("room.csv")

print("#2 Read Absorption Coefficient Data")
readAbsorpCoeffData("absorption.csv")

print("#3 Read Condition for Calculation")
readCalcCondData("control4sample2.csv")

print("#4 Calculate coefficients of Plane Equation")
calcPlaneEq()

print("#5 Create Sound Souce")
numberOfRays = createOmniSoundSource()

print("#6 Shoot off Sound Rays!!!!!!")

############################
### define local varies ####
############################
## propageted distance
distance = zeros((numberOfRays,1), "float64")

## number of reflection times
reflectionTimes = zeros((numberOfRays, 1), "int")

## maximum distance for calculation
limitDistance = totalTimeOfCalc * soundSpeed 

# energy of all sound particle in each 1/1 Octave band (125 to 4k Hz)
energy = ones((numberOfRays,6), "float64") / numberOfRays

## closs point coordinates
crossPoint = zeros((numberOfRays,3), "float64")

## temporary cross point coordinates
tmpCrossPoint = zeros((numberOfRays,3), "float64")

## minimum length to cross point
global minDistance

## index of reflection wall
crossWallIndex = zeros((numberOfRays,1), "int")

## life of sound particle
aliveOrDead = ones((numberOfRays,1), "bool")

#########################################
### 4.1 calculate reflection vector #####
#########################################
# @param rayIdx: Index of sound ray
#
def searchCrossWall(rayIdx):
    global minDistance
    minDistance = 0.0
    for wallIdx in range(numberOfWalls):
        if(isTowardWall( rayVector[rayIdx], planeEquation[wallIdx][0:3] ) == True ):
            tmpCrossPoint = calcCrossPoint(rayVector[rayIdx], 
                                           rayPosition[rayIdx], 
                                           planeEquation[wallIdx])
            if( norm(tmpCrossPoint) < 10.0**3.0 ):     # avoid math domain error
                if( isInsideWall( tmpCrossPoint, wallIdx ) == True ):
                    tmpDistance = \
                        calcPointToPointDistance(tmpCrossPoint, rayPosition[rayIdx])
                    if(minDistance == 0.0 or minDistance > tmpDistance):
                        minDistance = tmpDistance
                        crossWallIndex[rayIdx] = wallIdx
                        crossPoint[rayIdx] = tmpCrossPoint

##################################
### 4.2 draw sound particles #####
##################################
#
fig1 = matplotlib.pyplot.figure(1, figsize=(7,7))

if matplotlib.__version__ < '3.4':
    ax1 = Axes3D(fig1)
else:
    ax1 = Axes3D(fig1,auto_add_to_figure=False)
    fig1.add_axes(ax1)

def drawScat():
    x = rayPosition[0:numberOfRays-1,0]
    y = rayPosition[0:numberOfRays-1,1]
    z = rayPosition[0:numberOfRays-1,2]
    cc = reflectionTimes[0:numberOfRays-1]

    scat = ax1.scatter(x, y, z, c=cc, vmin=0, vmax=limitNumOfReflectTimes)
    oldScat = scat
    ax1.set_xlim3d( [min(point[:,0]), max(point[:,0])] )
    ax1.set_ylim3d( [min(point[:,1]), max(point[:,1])] )
    ax1.set_zlim3d( [min(point[:,2]), max(point[:,2])] )
    ax1.set_xlabel('x')
    ax1.set_ylabel('y')
    ax1.set_zlabel('z')
    matplotlib.pyplot.pause(0.1)  ## add this line for Python3
    matplotlib.pyplot.draw()
    ax1.collections.remove(oldScat)
    

###################################################
###    main loop
###################################################
drawScat()
#for k in range(int(totalTimeOfCalc/timeInterval)+1):
for k in range(5):
    if(sum(aliveOrDead) == 0):
        break
    for m in range(numberOfRays):
        #-------------------------------
        #  search cross wall 
        #-------------------------------
        leftDistance = timeInterval * soundSpeed
        while(leftDistance != 0.0):
            searchCrossWall(m)
            #-------------------------------
            # Do this ray hit the wall?
            #-------------------------------                
            if(minDistance <= leftDistance and aliveOrDead[m] == True):
                #------------------------------------------------
                # move to wall surface and refine sound ray properties     
                #------------------------------------------------
                rayVector[m] = calcReflectionVector(rayVector[m], 
                                      planeEquation[int(crossWallIndex[m])][0:3])
                rayPosition[m] = crossPoint[m]
                reflectionTimes[m] += 1
                if(reflectionTimes[m] == limitNumOfReflectTimes+1):
                    aliveOrDead[m] = False
                    rayPosition[m] = [1000, 1000, 1000]  ## go far away!!
                leftDistance -= minDistance
                minDistance = 0.0
            else:
                rayPosition[m] += rayVector[m] * leftDistance
                leftDistance = 0.0
                minDistance = 0.0  
    drawScat()
    

