#  ooo  oooo  oooo o   o     o    ooo   ooo  o   o  oooo ooooo o  ooo   oooo
# 0   0 0   0 0    00  0    0 0  0   0 0   0 0   0 0       0   0 0   0 0
# 0   0 oooo  oooo 0 0 0   0   0 0     0   0 0   0  oooo   0   0 0      oooo
# 0   0 o     0    0 o 0   ooooo 0   o 0   0 0   0      0  0   0 0   o      0
#  ooo  o     oooo o  0o  0     0 ooo   ooo   ooo  0oooo   0   0  ooo  0oooo
#
#  (  (  (  (  (  (  (  (  (  (  (  (  O  )  )  )  )  )  )  )  )  )  )  )  )
#
#    OpenAcoustics: The Open Source Collection for Computational Acoustics
#
#               Copyright (C) 2008-2009 The OpenAcoustics Team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

__author__ = "t_ishiduka <http://sourceforge.net.users.t_ishiduka>" 
__date__ = "2010/05/10"
__credits__ = "ISHIZUKA Takashi"
__version__ = "$Revision: 0 $"

# import modules 
from numpy import array, isinf, real, linspace, hstack, tile, sum
from scipy.special.orthogonal import p_roots

##Gauss-Legendre quadrature class
class GLquad:
    """ Gauss-Legendre quadrature class"""

    ## Initialize class
    #  @param self The self object
    #  @param low Lower limit of integration (default:-1.0)
    #  @param up Upper limit of integration (default:1.0)
    #  @param ord Order of quadrature (ord>0,default:4)
    #  @param ndiv Number of division whitch integration range is devided into (ndiv>0, default:1)
    def __init__(self, low=-1.0, up=1.0, ord=4, ndiv=1):
        """Initialize class"""
        
        self._low = low
        self._up = up

        self.setAcc(ord, ndiv)


    ## Set accuracy of quadrature
    #  @param self The self object
    #  @param ord Order of quadrature (ord>0,default:4)
    #  @param ndiv Number of division whitch integration range is divided into (ndiv>0, default:1)
    def setAcc(self, ord=4, ndiv=1):
        """Set accuracy of  quadrature"""

        if ord <= 0:
            raise "ord (order of quadrature) must be positive value"
        if ndiv <=0:
            raise "ndiv (number of division) must be positive value"

        self._ord = int(ord)
        self._ndiv = int(ndiv)

        self.setRange(self._low, self._up)


    ## Set range of integration
    #  @param self The self object
    #  @param low Lower limit of integration (default:-1.0)
    #  @param up Upper limit of integration (default:1.0)
    def setRange(self, low=-1.0, up=1.0):
        """Set range of integration"""

        if isinf(low) or isinf(up):
            raise "The quadrature is only for finite range"

        low = float(low)
        up = float(up)
        self._low = low
        self._up = up

        tmpx,tmpw = p_roots(self._ord)
        # normalize to range 0->1 
        tmpx = (real(tmpx) + 1.0) / 2.0
        tmpw = tmpw / 2.0

        # divide integration range
        divWidth = (up - low) / self._ndiv 
        divPoint = linspace(low,up,self._ndiv+1)
        
        # node points for each division
        xx = []
        for a in divPoint[:-1]:
            xx.append(divWidth * tmpx + a)

        # for entire range
        self._x = hstack(xx)

        # weighting values
        self._w = tile(divWidth * tmpw, self._ndiv)


    ## Execute quadrature
    #  @param self The self object
    #  @param func Integrand function (func must be vectorized)
    #  @param *arg arguments passed to integrand function  
    def quad(self, func, *arg):
        """ Execute quadrature"""

        return sum(self._w * func(self._x, *arg))


    ## Integration range
    def range(self):
        """integration range"""
        return [self._low,self._up]

    ## Accuracy of quadrature
    def acc(self):
        """Accuracy of quadrature"""
        return [self._ord, self._ndiv]

    ## Node points
    def x(self):
        """Evaluation points"""
        return self._x.copy()

    ## Weighting values
    def w(self):
        """Weighting values"""
        return self._w.copy()
