#  ooo  oooo  oooo o   o     o    ooo   ooo  o   o  oooo ooooo o  ooo   oooo
# 0   0 0   0 0    00  0    0 0  0   0 0   0 0   0 0       0   0 0   0 0
# 0   0 oooo  oooo 0 0 0   0   0 0     0   0 0   0  oooo   0   0 0      oooo
# 0   0 o     0    0 o 0   ooooo 0   o 0   0 0   0      0  0   0 0   o      0
#  ooo  o     oooo o  0o  0     0 ooo   ooo   ooo  0oooo   0   0  ooo  0oooo
#
#  (  (  (  (  (  (  (  (  (  (  (  (  O  )  )  )  )  )  )  )  )  )  )  )  )
#
#    OpenAcoustics: The Open Source Collection for Computational Acoustics
#
#               Copyright (C) 2008-2012 The OpenAcoustics Team
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


print('Import modules...\n\n')
# import modules 
from numpy import array, linalg, zeros, dot, pi, sqrt, ceil, linspace, meshgrid, vstack
from scipy import special
#from scipy.integrate import quadrature
from quadrature import GLquad
from openacoustics import gmsh
from scipy import mgrid
import pylab
import sys
from shapely.geometry import Point, Polygon

#====================================================================
#                        Definitions of functions
#=====================================================================
# Hankel functions
def h01(x):
    return special.j0(x) + 1j * special.y0(x)

def h11(x):
    return special.j1(x) + 1j * special.y1(x)

# Green function of the 2-dimensional free field
def GreenFunction(rp, rs, waveNumber):
    rp_rs = rp - rs
    r = sqrt(rp_rs[0] ** 2. + rp_rs[1] ** 2.)

    return (1j / 4.) * h01(r * waveNumber)

# Integrands of element integral for h and g
def integrand_h(xi, element, rp, waveNumber):
    xr = element['middlePoint'][0] + element['localAxis'][0] * xi
    yr = element['middlePoint'][1] + element['localAxis'][1] * xi
    xp_xr = rp[0] - xr
    yp_yr = rp[1] - yr
    r = sqrt(xp_xr ** 2. + yp_yr ** 2.)

    dr_dn = - (xp_xr * element['normalVector'][0] \
             + yp_yr * element['normalVector'][1]) / r

    return h11(r * waveNumber) * dr_dn

def integrand_g(xi, element, rp, waveNumber):
    xr = element['middlePoint'][0] + element['localAxis'][0] * xi
    yr = element['middlePoint'][1] + element['localAxis'][1] * xi
    xp_xr = rp[0] - xr
    yp_yr = rp[1] - yr
    r = sqrt(xp_xr ** 2. + yp_yr ** 2.)

    return h01(r * waveNumber)


#=====================================================================
#                              Main routine
#=====================================================================
#========================================================== Initialize 
#-- get a geometry file name
args = sys.argv

if (len(args) >= 2):
    geometryFile = args[1]
else:
    print('A geometry file name has to be given.')
    geometryFile = []


#-- sound speed -- assuming temperature = 15 in Celsius degree.
soundSpeed = 340.

#-- frequency --
frequency = 250.
angularFrequency = 2. * pi * frequency
waveNumber = angularFrequency / soundSpeed

#-- sources & observing points --
# temporarily load a geometry file to set the source point 
# and the range of the mesh grid

# create a mesher instance of GMsh2D class
mesher = gmsh.GMsh2D()
mesher.loadGeo(geometryFile, 10)
node = mesher.getNodes()

# source position
rs = node[ mesher.getPoints('S')[0]]
print('Source position : ', rs)
# remove the source position from node
node = node[:-1,:] 

# mesh grid of observing points
print('Creating mesh grid of observing points ...')
rmax = node.max(0);
rmin = node.min(0);

dr = 0.05;
cr = 0.025; # clearance gap from the boundary to the mesh grid
rmax = rmax - cr;
rmin = rmin + cr;
rpl = int(ceil((rmax[0]-rmin[0])/dr)) + 1
rpk = int(ceil((rmax[1]-rmin[1])/dr)) + 1
rx = linspace(rmin[0], rmax[0], rpl)
ry = linspace(rmax[1], rmin[1], rpk)
rpX, rpY = meshgrid(rx, ry)

print('  The number of observing points : ', rpk*rpl)

# make boundary polygon
boundary = Polygon(vstack([node, node[0,:]]))

#-- quadrature --
quad = GLquad(-1,1,6,2)

#===================================== Divide boundaries by using GMsh
print('Creating boundary elements...')

# set boundary conditions of lines
boundaryCondition = {'rigid':0., 'absorbing':1., 'soft':100000.}

# create a mesher instance of GMsh2D class
#mesher = gmsh.GMsh2D()

# create elements using GMsh and get nodes of elements
mesher.loadGeo(geometryFile, .1)
node = mesher.getNodes()

# initialize
element = []
# for each boundary contidion
for conditionName in boundaryCondition.keys():
    if mesher.getnameid(conditionName) != -1:
        # get node indexes defining elements
        nodeIndex = mesher.getLines(conditionName)

        # append elements and set r1, r2 and admittance
        for idx in nodeIndex:
            element.append({'r1':node[idx[0]], 'r2':node[idx[1]],
                'admittance':boundaryCondition[conditionName]})

nElement = len(element)

print('  The number of elements : ', nElement)

#-- element parameters --
for n in range(nElement):
    element[n]['middlePoint'] = (element[n]['r2'] + element[n]['r1']) / 2.
    r2_r1 = element[n]['r2'] - element[n]['r1']
    element[n]['localAxis'] = r2_r1 / 2.
    element[n]['length'] = sqrt(r2_r1[0] ** 2. + r2_r1[1] ** 2.)
    element[n]['normalVector'] = array([ r2_r1[1], - r2_r1[0] ]) \
            / element[n]['length']

#======================================== Build coefficient matrix 'A'
print('Carring out BEM ...')
print('  Building the coefficient matrix \'A\' ...')

A = zeros((nElement,nElement), 'complex64')

# for each boundary element
for m in range(nElement): 
    a = -1j * waveNumber * element[m]['length'] / (4. * 2.)
    b = waveNumber * element[m]['length'] \
            * element[m]['admittance'] / (4. * 2.)

    # for each middle point of element
    for l in range(nElement): 
        rl = element[l]['middlePoint']

        if l != m: # general component
#            h = a * quadrature(integrand_h, -1., 1., 
#                    args=(element[m], rl, waveNumber))[0]
#            g = b * quadrature(integrand_g, -1., 1., 
#                    args=(element[m], rl, waveNumber))[0]
            h = a * quad.quad(integrand_h, element[m], rl, waveNumber)
            g = b * quad.quad(integrand_g, element[m], rl, waveNumber)
            A[l,m] = h + g

        else: # diagonal component 
            v = special.itj0y0(waveNumber * element[m]['length'] / 2.)
            g = (element[l]['admittance'] / 2.) * complex(v[0], v[1])
            A[l,m] = 1./2. +  g

#======================================== Build right hand vector 'pd'
print('  Building the right hand vector \'pd\' ...')

pd = zeros((nElement,1), 'complex64')

for l in range(nElement):
    rl = element[l]['middlePoint']
    pd[l] = GreenFunction( rl, rs, waveNumber )

#==================================== Solve linear equation A * p = pd
print('  Solving the linear equation \'A * p = pd\' ...')

p = linalg.solve( A, pd )

#====================================== Calculate sound pressure at rp
print('  Calculating sound pressure at', rpk, '*', rpl, 'point(s) ...')

prp = zeros((rpk, rpl), 'complex64' )
Ar = zeros(nElement, 'complex64')

for k in range(rpk):
    for l in range(rpl):
        rp = array([rpX[k,l], rpY[k,l]])

        if Point(rp).within(boundary):   
            # check point rp is within the area surrounded by the boundary
            # see https://automating-gis-processes.github.io/CSC18/lessons/L4/point-in-polygon.html
            for m in range(nElement):
                a = -1j * waveNumber * element[m]['length'] / (4. * 2.)
                b = waveNumber * element[m]['length'] * \
                        element[m]['admittance'] / (4. * 2.)

#            h = a * quadrature(integrand_h, -1., 1., 
#                    args=(element[m], rp, waveNumber))[0]
#            g = b * quadrature(integrand_g, -1., 1., 
#                    args=(element[m], rp, waveNumber))[0]
                h = a * quad.quad(integrand_h, element[m], rp, waveNumber)
                g = b * quad.quad(integrand_g, element[m], rp, waveNumber)
                Ar[m] = h + g

            prp[k,l] = GreenFunction(rp, rs, waveNumber) - dot(Ar, p)[0]
        else:
            prp[k,l] = float('nan')


#======================================================== Show results
print('Creating a contour plot of abs(Prp) ...')

# plot contour
pylab.contourf(rpX, rpY, abs(prp), 32, cmap=pylab.cm.rainbow)

# plot boundary elements
# If the number of elements is too large, 
# you should comment out these lines
for conditionName in boundaryCondition.keys():
    if mesher.getnameid(conditionName) != -1:
        nodeIndex = mesher.getLines(conditionName)
        if boundaryCondition[conditionName] == 0.0:
            lineStyle = '-bx'
        elif boundaryCondition[conditionName] == 1.0:
            lineStyle = '--mx'
        elif boundaryCondition[conditionName] > 1.0:
            lineStyle = '-cx'
        else:
            lineStyle = '--yx'

        for idx in nodeIndex:
            pylab.plot(node[idx][:,0], node[idx][:,1], lineStyle)

#nodeIndex = mesher.getLines('absorbing')
#for idx in nodeIndex:
#    pylab.plot(node[idx][:,0], node[idx][:,1], '--rx')
#

# plot source position
pylab.plot(array([rs[0]]), array([rs[1]]), 'ro', markeredgecolor='white')

# arrange graph view
pylab.axis('equal')
pylab.xlabel('x [m]')
pylab.ylabel('y [m]')
pylab.title('Absolute values of sound pressure at 250 Hz')
pylab.colorbar()

# save and show graph
print('\nClose the figure window to exit this code.')
pylab.savefig('bem_'+geometryFile[:-4]+'_at250Hz.png')
pylab.show()

# End

